"""
Demonstrates item creation as well as bodypart / clothing upload
"""

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.templates import WearableType, Permissions
from hippolyzer.lib.base.wearables import Wearable, VISUAL_PARAMS
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class ShapeCreatorAddon(BaseAddon):
    @handle_command()
    async def create_shape(self, session: Session, region: ProxiedRegion):
        """Make a shape with pre-set parameters and place it in the body parts folder"""

        wearable = Wearable.make_default(WearableType.SHAPE)
        # Max out the jaw jut param
        jaw_param = VISUAL_PARAMS.by_name("Jaw Jut")
        wearable.parameters[jaw_param.id] = jaw_param.value_max
        wearable.name = "Cool Shape"

        # A unique transaction ID is needed to tie the item creation to the following asset upload.
        transaction_id = UUID.random()
        item = await session.inventory.create_item(
            UUID.ZERO,  # This will place it in the default folder for the type
            name=wearable.name,
            type=wearable.wearable_type.asset_type,
            inv_type=wearable.wearable_type.asset_type.inventory_type,
            wearable_type=wearable.wearable_type,
            next_mask=Permissions.MOVE | Permissions.MODIFY | Permissions.COPY | Permissions.TRANSFER,
            transaction_id=transaction_id,
        )
        print(f"Created {item!r}")
        await region.xfer_manager.upload_asset(
            wearable.wearable_type.asset_type,
            wearable.to_str(),
            transaction_id=transaction_id,
        )


addons = [ShapeCreatorAddon()]



"""
Example of how to control a viewer over LEAP

Must launch the viewer with `outleap-agent` LEAP script.
See https://github.com/SaladDais/outleap/ for more info on LEAP / outleap.
"""

import outleap
from outleap.scripts.inspector import LEAPInspectorGUI

from hippolyzer.lib.proxy.addon_utils import send_chat, BaseAddon, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager


# Path found using `outleap-inspector`
FPS_PATH = outleap.UIPath("/main_view/menu_stack/status_bar_container/status/time_and_media_bg/FPSText")


class LEAPExampleAddon(BaseAddon):
    async def handle_leap_client_added(self, session_manager: SessionManager, leap_client: outleap.LEAPClient):
        # You can do things as soon as the LEAP client connects, like if you want to automate
        # login or whatever.
        viewer_control_api = outleap.LLViewerControlAPI(leap_client)
        # Ask for a config value and print it in the viewer logs
        print(await viewer_control_api.get("Global", "StatsPilotFile"))

    @handle_command()
    async def show_ui_inspector(self, session: Session, _region: ProxiedRegion):
        """Spawn a GUI for inspecting the UI state"""
        if not session.leap_client:
            show_message("No LEAP client connected?")
            return
        LEAPInspectorGUI(session.leap_client).show()

    @handle_command()
    async def say_fps(self, session: Session, _region: ProxiedRegion):
        """Say your current FPS in chat"""
        if not session.leap_client:
            show_message("No LEAP client connected?")
            return

        window_api = outleap.LLWindowAPI(session.leap_client)
        fps = (await window_api.get_info(path=FPS_PATH))['value']

        send_chat(f"LEAP says I'm running at {fps} FPS!")


addons = [LEAPExampleAddon()]



from __future__ import annotations

import asyncio
import dataclasses
import gzip
import itertools
import logging
from pathlib import Path
from typing import Union, List, Tuple, Set, Sequence, Dict, TYPE_CHECKING

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.inventory import InventoryModel, InventoryCategory, InventoryItem, InventoryNodeBase
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.templates import AssetType, FolderType, InventoryType, Permissions
from hippolyzer.lib.base.templates import WearableType

if TYPE_CHECKING:
    from hippolyzer.lib.client.state import BaseClientSession

LOG = logging.getLogger(__name__)


class CannotMoveError(Exception):
    def __init__(self):
        pass


def _get_node_id(node_or_id: InventoryNodeBase | UUID) -> UUID:
    if isinstance(node_or_id, UUID):
        return node_or_id
    return node_or_id.node_id


class InventoryManager:
    def __init__(self, session: BaseClientSession):
        self._session = session
        self.model: InventoryModel = InventoryModel()
        self._load_skeleton()
        self._session.message_handler.subscribe("BulkUpdateInventory", self._handle_bulk_update_inventory)
        self._session.message_handler.subscribe("UpdateCreateInventoryItem", self._handle_update_create_inventory_item)
        self._session.message_handler.subscribe("RemoveInventoryItem", self._handle_remove_inventory_item)
        self._session.message_handler.subscribe("RemoveInventoryObjects", self._handle_remove_inventory_objects)
        self._session.message_handler.subscribe("RemoveInventoryFolder", self._handle_remove_inventory_folder)
        self._session.message_handler.subscribe("MoveInventoryItem", self._handle_move_inventory_item)
        self._session.message_handler.subscribe("MoveInventoryFolder", self._handle_move_inventory_folder)

    def _load_skeleton(self):
        assert not self.model.nodes
        skel_cats: List[dict] = self._session.login_data.get('inventory-skeleton', [])
        for skel_cat in skel_cats:
            self.model.add(InventoryCategory(
                name=skel_cat["name"],
                cat_id=UUID(skel_cat["folder_id"]),
                parent_id=UUID(skel_cat["parent_id"]),
                # Don't use the version from the skeleton, this flags the inventory as needing
                # completion from the inventory cache. This matches indra's behavior.
                version=InventoryCategory.VERSION_NONE,
                type=AssetType.CATEGORY,
                pref_type=FolderType(skel_cat.get("type_default", FolderType.NONE)),
                owner_id=self._session.agent_id,
            ))

    def load_cache(self, path: Union[str, Path]):
        # Per indra, rough flow for loading inv on login is:
        # 1. Look at inventory skeleton from login response
        # 2. Pre-populate model with categories from the skeleton, including their versions
        # 3. Read the inventory cache, tracking categories and items separately
        # 4. Walk the list of categories in our cache. If the cat exists in the skeleton and the versions
        #    match, then we may load the category and its descendants from cache.
        # 5. Any categories in the skeleton but not in the cache, or those with mismatched versions must be fetched.
        #    The viewer does this by setting the local version of the cats to -1 and forcing a descendent fetch
        #    over AIS.
        #
        # By the time you call this function call, you should have already loaded the inventory skeleton
        # into the model set its inventory category versions to VERSION_NONE.

        skel_cats: List[dict] = self._session.login_data['inventory-skeleton']
        # UUID -> version map for inventory skeleton
        skel_versions = {UUID(cat["folder_id"]): cat["version"] for cat in skel_cats}
        LOG.info(f"Parsing inv cache at {path}")
        cached_categories, cached_items = self._parse_cache(path)
        LOG.info(f"Done parsing inv cache at {path}")
        loaded_cat_ids: Set[UUID] = set()

        for cached_cat in cached_categories:
            existing_cat: InventoryCategory = self.model.get(cached_cat.cat_id)  # noqa
            # Don't clobber an existing cat unless it just has a placeholder version,
            # maybe from loading the skeleton?
            if existing_cat and existing_cat.version != InventoryCategory.VERSION_NONE:
                continue
            # Cached cat isn't the same as what the inv server says it should be, can't use it.
            if cached_cat.version != skel_versions.get(cached_cat.cat_id):
                continue
            # Update any existing category in-place, or add if not present
            self.model.upsert(cached_cat)
            # Any items in this category in our cache file are usable and should be added
            loaded_cat_ids.add(cached_cat.cat_id)

        for cached_item in cached_items:
            # The skeleton doesn't have any items, so if we run into any items they should be exactly the
            # same as what we're trying to add. No point clobbering.
            if cached_item.item_id in self.model:
                continue
            # The parent category didn't have a cache hit against the inventory skeleton, can't add!
            # We don't even know if this item would be in the current version of its parent cat!
            if cached_item.parent_id not in loaded_cat_ids:
                continue
            self.model.add(cached_item)

        self.model.flag_if_dirty()

    def _parse_cache(self, path: Union[str, Path]) -> Tuple[List[InventoryCategory], List[InventoryItem]]:
        """Warning, may be incredibly slow due to llsd.parse_notation() behavior"""
        categories: List[InventoryCategory] = []
        items: List[InventoryItem] = []
        # Parse our cached items and categories out of the compressed inventory cache
        first_line = True
        with gzip.open(path, "rb") as f:
            # Line-delimited LLSD notation!
            for line in f.readlines():
                # TODO: Parsing of invcache is dominated by `parse_notation()`. It's stupidly inefficient.
                # TODO: sniff out binary LLSD invcaches
                node_llsd = llsd.parse_notation(line)
                if first_line:
                    # First line is the file header
                    first_line = False
                    if node_llsd['inv_cache_version'] not in (2, 3):
                        raise ValueError(f"Unknown cache version: {node_llsd!r}")
                    continue

                if InventoryCategory.ID_ATTR in node_llsd:
                    if (cat_node := InventoryCategory.from_llsd(node_llsd)) is not None:
                        categories.append(cat_node)
                elif InventoryItem.ID_ATTR in node_llsd:
                    if (item_node := InventoryItem.from_llsd(node_llsd)) is not None:
                        items.append(item_node)
                else:
                    LOG.warning(f"Unknown node type in inv cache: {node_llsd!r}")
        return categories, items

    def _handle_bulk_update_inventory(self, msg: Message):
        any_cats = False
        for folder_block in msg.get_blocks("FolderData", ()):
            if folder_block["FolderID"] == UUID.ZERO:
                continue
            any_cats = True
            self.model.upsert(
                InventoryCategory.from_folder_data(folder_block),
                # Don't clobber version, we only want to fetch the folder if it's new
                # and hasn't just moved.
                update_fields={"parent_id", "name", "pref_type"},
            )
        for item_block in msg.get_blocks("ItemData", ()):
            if item_block["ItemID"] == UUID.ZERO:
                continue
            self.model.upsert(InventoryItem.from_inventory_data(item_block))

        if any_cats:
            self.model.flag_if_dirty()

    def _validate_recipient(self, recipient: UUID):
        if self._session.agent_id != recipient:
            raise ValueError(f"AgentID Mismatch {self._session.agent_id} != {recipient}")

    def _handle_update_create_inventory_item(self, msg: Message):
        self._validate_recipient(msg["AgentData"]["AgentID"])
        for inventory_block in msg["InventoryData"]:
            self.model.upsert(InventoryItem.from_inventory_data(inventory_block))

    def _handle_remove_inventory_item(self, msg: Message):
        self._validate_recipient(msg["AgentData"]["AgentID"])
        for inventory_block in msg["InventoryData"]:
            node = self.model.get(inventory_block["ItemID"])
            if node:
                self.model.unlink(node)

    def _handle_remove_inventory_folder(self, msg: Message):
        self._validate_recipient(msg["AgentData"]["AgentID"])
        for folder_block in msg["FolderData"]:
            node = self.model.get(folder_block["FolderID"])
            if node:
                self.model.unlink(node)

    def _handle_remove_inventory_objects(self, msg: Message):
        self._validate_recipient(msg["AgentData"]["AgentID"])
        for item_block in msg.get_blocks("ItemData", []):
            node = self.model.get(item_block["ItemID"])
            if node:
                self.model.unlink(node)
        for folder_block in msg.get_blocks("FolderData", []):
            node = self.model.get(folder_block["FolderID"])
            if node:
                self.model.unlink(node)

    def _handle_move_inventory_item(self, msg: Message):
        for inventory_block in msg["InventoryData"]:
            node = self.model.get(inventory_block["ItemID"])
            if not node:
                LOG.warning(f"Missing inventory item {inventory_block['ItemID']}")
                continue
            if inventory_block["NewName"]:
                node.name = str(inventory_block["NewName"])
            node.parent_id = inventory_block['FolderID']

    def _handle_move_inventory_folder(self, msg: Message):
        for inventory_block in msg["InventoryData"]:
            node = self.model.get(inventory_block["FolderID"])
            if not node:
                LOG.warning(f"Missing inventory folder {inventory_block['FolderID']}")
                continue
            node.parent_id = inventory_block['ParentID']

    def process_aisv3_response(self, payload: dict):
        if "name" in payload:
            # Just a rough guess. Assume this response is updating something if there's
            # a "name" key.
            if InventoryCategory.ID_ATTR_AIS in payload:
                if (cat_node := InventoryCategory.from_llsd(payload, flavor="ais")) is not None:
                    self.model.upsert(cat_node)
            elif InventoryItem.ID_ATTR in payload:
                if (item_node := InventoryItem.from_llsd(payload, flavor="ais")) is not None:
                    self.model.upsert(item_node)
            else:
                LOG.warning(f"Unknown node type in AIS payload: {payload!r}")

        # Parse the embedded stuff
        embedded_dict = payload.get("_embedded", {})
        for category_llsd in embedded_dict.get("categories", {}).values():
            self.model.upsert(InventoryCategory.from_llsd(category_llsd, flavor="ais"))
        for item_llsd in embedded_dict.get("items", {}).values():
            self.model.upsert(InventoryItem.from_llsd(item_llsd, flavor="ais"))
        for link_llsd in embedded_dict.get("links", {}).values():
            self.model.upsert(InventoryItem.from_llsd(link_llsd, flavor="ais"))

        for cat_id, version in payload.get("_updated_category_versions", {}).items():
            # The key will be a string, so convert to UUID first
            cat_node = self.model.get_category(UUID(cat_id))
            cat_node.version = version

        # Get rid of anything we were asked to
        for node_id in itertools.chain(
                payload.get("_broken_links_removed", ()),
                payload.get("_removed_items", ()),
                payload.get("_category_items_removed", ()),
                payload.get("_categories_removed", ()),
        ):
            node = self.model.get(node_id)
            if node:
                # Presumably this list is exhaustive, so don't unlink children.
                self.model.unlink(node, single_only=True)

    async def make_ais_request(
            self,
            method: str,
            path: str,
            params: dict,
            payload: dict | Sequence | dataclasses.MISSING = dataclasses.MISSING,
    ) -> dict:
        caps_client = self._session.main_region.caps_client
        async with caps_client.request(method, "InventoryAPIv3", path=path, params=params, llsd=payload) as resp:
            if resp.ok or resp.status == 400:
                data = await resp.read_llsd()
                if err_desc := data.get("error_description", ""):
                    err_desc: str
                    if err_desc.startswith("Cannot change parent_id."):
                        raise CannotMoveError()
                resp.raise_for_status()
                self.process_aisv3_response(data)
            else:
                resp.raise_for_status()

            return data

    async def create_folder(
            self,
            parent: InventoryCategory | UUID,
            name: str,
            pref_type: int = AssetType.NONE,
            cat_id: UUID | None = None
    ) -> InventoryCategory:
        parent_id = _get_node_id(parent)

        payload = {
            "categories": [
                {
                    "category_id": cat_id,
                    "name": name,
                    "type_default": pref_type,
                    "parent_id": parent_id
                }
            ]
        }
        data = await self.make_ais_request("POST", f"/category/{parent_id}", {"tid": UUID.random()}, payload)
        return self.model.get_category(data["_created_categories"][0])

    async def create_item(
            self,
            parent: UUID | InventoryCategory,
            name: str,
            type: AssetType,
            inv_type: InventoryType,
            wearable_type: WearableType,
            transaction_id: UUID,
            next_mask: int | Permissions = 0x0008e000,
            description: str = '',
    ) -> InventoryItem:
        parent_id = _get_node_id(parent)

        with self._session.main_region.message_handler.subscribe_async(
                ("UpdateCreateInventoryItem",),
                predicate=lambda x: x["AgentData"]["TransactionID"] == transaction_id,
                take=False,
        ) as get_msg:
            await self._session.main_region.circuit.send_reliable(
                Message(
                    'CreateInventoryItem',
                    Block('AgentData', AgentID=self._session.agent_id, SessionID=self._session.id),
                    Block(
                        'InventoryBlock',
                        CallbackID=0,
                        FolderID=parent_id,
                        TransactionID=transaction_id,
                        NextOwnerMask=next_mask,
                        Type=type,
                        InvType=inv_type,
                        WearableType=wearable_type,
                        Name=name,
                        Description=description,
                    )
                )
            )
            msg = await asyncio.wait_for(get_msg(), 5.0)
            # We assume that _handle_update_create_inventory_item() has already been called internally
            # by the time that the `await` returns given asyncio scheduling
            return self.model.get_item(msg["InventoryData"]["ItemID"])

    async def move(self, node: InventoryNodeBase, new_parent: UUID | InventoryCategory) -> None:
        # AIS error messages suggest using the MOVE HTTP method instead of setting a new parent
        # via PATCH. MOVE is not implemented in AIS. Instead, we do what the viewer does and use
        # legacy UDP messages for reparenting things
        new_parent = _get_node_id(new_parent)

        msg = Message(
            "MoveInventoryFolder",
            Block("AgentData", AgentID=self._session.agent_id, SessionID=self._session.id, Stamp=0),
        )

        if isinstance(node, InventoryItem):
            msg.add_block(Block("InventoryData", ItemID=node.node_id, FolderID=new_parent, NewName=b''))
        else:
            msg.add_block(Block("InventoryData", FolderID=node.node_id, ParentID=new_parent))

        # No message to say if this even succeeded. Great.
        # TODO: probably need to update category versions for both source and target
        await self._session.main_region.circuit.send_reliable(msg)
        node.parent_id = new_parent

    async def copy(self, node: InventoryNodeBase, destination: UUID | InventoryCategory, contents: bool = True)\
            -> InventoryItem | InventoryCategory:
        destination = _get_node_id(destination)
        if isinstance(node, InventoryItem):
            with self._session.main_region.message_handler.subscribe_async(
                    ("BulkUpdateInventory",),
                    # Not ideal, but there doesn't seem to be an easy way to determine the transaction ID,
                    # and using the callback ID seems a bit crap.
                    predicate=lambda x: x["ItemData"]["Name"] == node.name,
                    take=False,
            ) as get_msg:
                await self._session.main_region.circuit.send_reliable(Message(
                    'CopyInventoryItem',
                    Block('AgentData', AgentID=self._session.agent_id, SessionID=self._session.id),
                    Block(
                        'InventoryData',
                        CallbackID=0,
                        OldAgentID=self._session.agent_id,
                        OldItemID=node.item_id,
                        NewFolderID=destination,
                        NewName=b''
                    )
                ))
                msg = await asyncio.wait_for(get_msg(), 5.0)
                # BulkInventoryUpdate message may not have already been handled internally, do it manually.
                self._handle_bulk_update_inventory(msg)

                # Now pull the item out of the inventory
                new_item = self.model.get(msg["ItemData"]["ItemID"])
                assert new_item is not None
                return new_item  # type: ignore
        elif isinstance(node, InventoryCategory):
            # Keep a list of the original descendents in case we're copy a folder within itself
            to_copy = list(node.descendents)
            # There's not really any way to "copy" a category, we just create a new one with the same properties.
            new_cat = await self.create_folder(destination, node.name, node.pref_type)
            if contents:
                cat_lookup: Dict[UUID, UUID] = {node.node_id: new_cat.node_id}
                # Recreate the category hierarchy first, keeping note of the new category IDs.
                for node in to_copy:
                    if isinstance(node, InventoryCategory):
                        new_parent = cat_lookup[node.parent_id]
                        cat_lookup[node.node_id] = (await self.copy(node, new_parent, contents=False)).node_id
                # Items have to be explicitly copied individually
                for node in to_copy:
                    if isinstance(node, InventoryItem):
                        new_parent = cat_lookup[node.parent_id]
                        await self.copy(node, new_parent, contents=False)
            return new_cat
        else:
            raise ValueError(f"Unknown node type: {node!r}")

    async def update(self, node: InventoryNodeBase, data: dict) -> None:
        path = f"/category/{node.node_id}"
        if isinstance(node, InventoryItem):
            path = f"/item/{node.node_id}"
        await self.make_ais_request("PATCH", path, {}, data)



"""
Control a puppetry-enabled viewer and make your neck spin like crazy

It currently requires a custom rebased Firestorm with puppetry applied on top,
and patches applied on top to make startup LEAP scripts be treated as puppetry modules.
Basically, you probably don't want to use this yet. But hey, Puppetry is still only
on the beta grid anyway.
"""
import asyncio
import enum
import logging
import math
from typing import *

import outleap

from hippolyzer.lib.base.datatypes import Quaternion
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty
from hippolyzer.lib.proxy.sessions import Session

LOG = logging.getLogger(__name__)


class BodyPartMask(enum.IntFlag):
    """Which joints to send the viewer as part of "move" puppetry command"""
    HEAD = 1 << 0
    FACE = 1 << 1
    LHAND = 1 << 2
    RHAND = 1 << 3
    FINGERS = 1 << 4


def register_puppetry_command(func: Callable[[dict], Awaitable[None]]):
    """Register a method as handling inbound puppetry commands from the viewer"""
    func._puppetry_command = True
    return func


class PuppetryExampleAddon(BaseAddon):
    server_skeleton: Dict[str, Dict[str, Any]] = SessionProperty(dict)
    camera_num: int = SessionProperty(0)
    parts_active: BodyPartMask = SessionProperty(lambda: BodyPartMask(0x1F))
    puppetry_api: Optional[outleap.LLPuppetryAPI] = SessionProperty(None)
    leap_client: Optional[outleap.LEAPClient] = SessionProperty(None)

    def handle_session_init(self, session: Session):
        if not session.leap_client:
            return
        self.puppetry_api = outleap.LLPuppetryAPI(session.leap_client)
        self.leap_client = session.leap_client
        self._schedule_task(self._serve())
        self._schedule_task(self._exorcist(session))

    @register_puppetry_command
    async def enable_parts(self, args: dict):
        if (new_mask := args.get("parts_mask")) is not None:
            self.parts_active = BodyPartMask(new_mask)

    @register_puppetry_command
    async def set_camera(self, args: dict):
        if (camera_num := args.get("camera_num")) is not None:
            self.camera_num = camera_num

    @register_puppetry_command
    async def stop(self, _args: dict):
        LOG.info("Viewer asked us to stop puppetry")

    @register_puppetry_command
    async def log(self, _args: dict):
        # Intentionally ignored, we don't care about things the viewer
        # asked us to log
        pass

    @register_puppetry_command
    async def set_skeleton(self, args: dict):
        # Don't really care about what the viewer thinks the view of the skeleton is.
        # Just log store it.
        self.server_skeleton = args

    async def _serve(self):
        """Handle inbound puppetry commands from viewer in a loop"""
        async with self.leap_client.listen_scoped("puppetry.controller") as listener:
            while True:
                msg = await listener.get()
                cmd = msg["command"]
                handler = getattr(self, cmd, None)
                if handler is None or not hasattr(handler, "_puppetry_command"):
                    LOG.warning(f"Unknown puppetry command {cmd!r}: {msg!r}")
                    continue
                await handler(msg.get("args", {}))

    async def _exorcist(self, session):
        """Do the Linda Blair thing with your neck"""
        spin_rad = 0.0
        while True:
            await asyncio.sleep(0.05)
            if not session.main_region:
                continue
            # Wrap spin_rad around if necessary
            while spin_rad > math.pi:
                spin_rad -= math.pi * 2

            # LEAP wants rot as a quaternion with just the imaginary parts.
            neck_rot = Quaternion.from_euler(0, 0, spin_rad).data(3)
            self.puppetry_api.move({
                "mNeck": {"no_constraint": True, "local_rot": neck_rot},
            })
            spin_rad += math.pi / 25


addons = [PuppetryExampleAddon()]



"""
Addon demonstrating a Qt GUI, use of the object manager and associated addon hooks

Displays a list of all objects that are mostly blue on at least one face based
on prim colors.
"""
from __future__ import annotations

import asyncio
import enum
import os.path
from typing import *

from PySide6 import QtCore, QtGui, QtWidgets

from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.objects import Object
from hippolyzer.lib.base.ui_helpers import loadUi
from hippolyzer.lib.base.templates import PCode
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.base.network.transport import Direction
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session
from hippolyzer.lib.proxy.task_scheduler import TaskLifeScope


def _is_color_blueish(color: bytes) -> bool:
    # Eh this is pretty transparent.
    if color[3] < 128:
        return False

    # pretty low value, more black than anything
    if color[2] < 50:
        return False

    # Blue channel makes up at least 70% of the value
    return (color[2] / sum(color[:3])) > 0.7


def _is_object_blueish(obj: Object):
    if obj.PCode != PCode.PRIMITIVE:
        return False
    for color in obj.TextureEntry.Color.values():
        if _is_color_blueish(color):
            return True
    return False


class BlueishObjectListGUIAddon(BaseAddon):
    blueish_model: Optional[BlueishObjectModel] = SessionProperty(None)

    # Cancel the coroutine associated with this command if the region, session or addon
    # changes for any reason. Only one allowed at once across all sessions.
    @handle_command(
        single_instance=True,
        lifetime=TaskLifeScope.SESSION | TaskLifeScope.REGION | TaskLifeScope.ADDON
    )
    async def track_blueish(self, session: Session, region: ProxiedRegion):
        """Open a window that tracks blueish objects in the region"""
        parent = AddonManager.UI.main_window_handle()
        if parent is None:
            raise RuntimeError("Must be run under the GUI proxy")

        win = BlueishObjectWindow(parent, session)
        win.objectHighlightClicked.connect(self._highlight_object)  # type: ignore
        win.objectTeleportClicked.connect(self._teleport_to_object)  # type: ignore
        win.show()
        try:
            self.blueish_model = win.model
            self._scan_all_objects(session, region)
            await win.closing
            self.blueish_model = None
        except:
            # Task got killed or something exploded, close the window ourselves
            self.blueish_model = None
            win.close()
            raise

    def _highlight_object(self, session: Session, obj: Object):
        session.main_region.circuit.send(Message(
            "ForceObjectSelect",
            Block("Header", ResetList=False),
            Block("Data", LocalID=obj.LocalID),
            direction=Direction.IN,
        ))

    def _teleport_to_object(self, session: Session, obj: Object):
        session.main_region.circuit.send(Message(
            "TeleportLocationRequest",
            Block("AgentData", AgentID=session.agent_id, SessionID=session.id),
            Block(
                "Info",
                RegionHandle=session.main_region.handle,
                Position=obj.RegionPosition,
                LookAt=Vector3(0.0, 0.0, 0.0)
            ),
        ))

    def _scan_all_objects(self, _session: Session, region: ProxiedRegion):
        self.blueish_model.clear()

        for obj in region.objects.all_objects:
            if _is_object_blueish(obj):
                self.blueish_model.addObject(obj)

        obj_list = self.blueish_model.objects
        region.objects.request_object_properties([o for o in obj_list if o.Name is None])

        # Make sure we request any objects we didn't know about before,
        # they'll get picked up in the update handler.
        region.objects.request_missing_objects()

    def handle_object_updated(self, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        if self.blueish_model is None:
            return

        if _is_object_blueish(obj):
            if obj not in self.blueish_model:
                if obj.Name is None:
                    region.objects.request_object_properties(obj)
                self.blueish_model.addObject(obj)
            else:
                # mark the object as updated in the model,
                # fields may have changed.
                self.blueish_model.updateObject(obj)
        else:
            if obj in self.blueish_model:
                self.blueish_model.removeObject(obj)

    def handle_object_killed(self, session: Session, region: ProxiedRegion, obj: Object):
        if self.blueish_model is None:
            return
        if obj in self.blueish_model:
            self.blueish_model.removeObject(obj)


class BlueishModelHeader(enum.IntEnum):
    Name = 0
    Position = enum.auto()


class BlueishObjectModel(QtCore.QAbstractTableModel):
    def __init__(self, parent):
        super().__init__(parent)
        self.objects: List[Object] = []

    def __contains__(self, item):
        return item in self.objects

    def addObject(self, obj: Object):
        if obj in self.objects:
            self.updateObject(obj)
            return

        num_objs = len(self.objects)
        self.beginInsertRows(QtCore.QModelIndex(), num_objs, num_objs)
        self.objects.append(obj)
        self.endInsertRows()

    def removeObject(self, obj: Object):
        try:
            obj_idx = self.objects.index(obj)
        except ValueError:
            return

        self.beginRemoveRows(QtCore.QModelIndex(), obj_idx, obj_idx)
        self.objects.remove(obj)
        self.endRemoveRows()

    def updateObject(self, obj: Object):
        try:
            obj_idx = self.objects.index(obj)
        except ValueError:
            return
        top_left = self.createIndex(obj_idx, 1)
        bottom_right = self.createIndex(obj_idx, self.columnCount())
        self.dataChanged.emit(top_left, bottom_right)

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.objects)

    def columnCount(self, parent: QtCore.QModelIndex = None) -> int:
        return len(BlueishModelHeader)

    def data(self, index, role=None):
        if not index.isValid():
            return None
        obj = self.objects[index.row()]
        if role == QtCore.Qt.UserRole:
            return obj
        if role != QtCore.Qt.DisplayRole:
            return None

        col = index.column()
        val = None

        if col == BlueishModelHeader.Name:
            val = obj.Name or ""
        elif col == BlueishModelHeader.Position:
            try:
                val = str(obj.RegionPosition)
            except ValueError:
                # If the object is orphaned we may not be able to figure
                # out the region pos
                val = "Unknown"

        return val

    def headerData(self, col, orientation, role=None):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return BlueishModelHeader(col).name

    def clear(self):
        self.beginResetModel()
        self.objects = []
        self.endResetModel()


BLUEISH_UI_PATH = os.path.join(os.path.dirname(__file__), "blueish_object_list.ui")


class BlueishObjectWindow(QtWidgets.QMainWindow):
    objectHighlightClicked = QtCore.Signal(Session, Object)
    objectTeleportClicked = QtCore.Signal(Session, Object)

    tableView: QtWidgets.QTableView

    def __init__(self, parent, session: Session):
        self.closing = asyncio.Future()
        super().__init__(parent=parent)
        loadUi(BLUEISH_UI_PATH, self)
        self.model = BlueishObjectModel(self)
        self.session = session
        self.tableView.setModel(self.model)
        self.tableView.horizontalHeader().resizeSection(BlueishModelHeader.Name, 150)
        self.tableView.horizontalHeader().setStretchLastSection(True)
        self.buttonHighlight.clicked.connect(self._highlightClicked)
        self.buttonTeleport.clicked.connect(self._teleportClicked)

    def closeEvent(self, event: QtGui.QCloseEvent):
        if not self.closing.done():
            self.closing.set_result(True)
        super().closeEvent(event)

    def _highlightClicked(self):
        self._emitForSelectedObject(self.objectHighlightClicked)

    def _teleportClicked(self):
        self._emitForSelectedObject(self.objectTeleportClicked)

    def _emitForSelectedObject(self, signal: QtCore.Signal):
        object_indexes = self.tableView.selectionModel().selectedIndexes()
        if not object_indexes:
            return
        obj = object_indexes[0].data(QtCore.Qt.UserRole)
        signal.emit(self.session, obj)  # type: ignore


addons = [BlueishObjectListGUIAddon()]



from typing import Mapping, Optional

import multidict

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.network.caps_client import CapsClient
from hippolyzer.lib.base.test_utils import MockHandlingCircuit
from hippolyzer.lib.client.hippo_client import ClientSettings
from hippolyzer.lib.client.object_manager import ClientWorldObjectManager
from hippolyzer.lib.client.state import BaseClientRegion, BaseClientSession, BaseClientSessionManager


class MockClientRegion(BaseClientRegion):
    def __init__(self, caps_urls: Optional[dict] = None):
        super().__init__()
        self.handle = None
        self.circuit_addr = ("127.0.0.1", 1)
        self.message_handler: MessageHandler[Message, str] = MessageHandler(take_by_default=False)
        self.circuit = MockHandlingCircuit(self.message_handler)
        self._name = "Test"
        self.cap_urls = multidict.MultiDict()
        if caps_urls:
            self.cap_urls.update(caps_urls)
        self.caps_client = CapsClient(self.cap_urls)

    def session(self):
        return MockClientSession(UUID.ZERO, UUID.ZERO, UUID.ZERO, 0, None)

    def update_caps(self, caps: Mapping[str, str]) -> None:
        pass


class MockClientSession(BaseClientSession):
    def __init__(self, id, secure_session_id, agent_id, circuit_code,
                 session_manager: Optional[BaseClientSessionManager]):
        super().__init__(id, secure_session_id, agent_id, circuit_code, session_manager)
        self.objects = ClientWorldObjectManager(self, ClientSettings(), None)



import asyncio
import datetime as dt
import functools
import logging
from typing import *

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.helpers import get_mtime, create_logged_task
from hippolyzer.lib.client.inventory_manager import InventoryManager
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.viewer_settings import iter_viewer_cache_dirs
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.inventory import InventoryCategory, InventoryNodeBase
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.inventory import InventoryItem
from hippolyzer.lib.base.templates import AssetType, InventoryType, WearableType, Permissions
from hippolyzer.lib.base.network.transport import Direction

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.sessions import Session


LOG = logging.getLogger(__name__)


class ProxyInventoryManager(InventoryManager):
    _session: "Session"

    def __init__(self, session: "Session"):
        # These handlers all need their processing deferred until the cache has been loaded.
        # Since cache is loaded asynchronously, the viewer may get ahead of us due to parsing
        # the cache faster and start requesting inventory details we can't do anything with yet.
        self._handle_update_create_inventory_item = self._wrap_with_cache_defer(
            self._handle_update_create_inventory_item
        )
        self._handle_remove_inventory_item = self._wrap_with_cache_defer(
            self._handle_remove_inventory_item
        )
        self._handle_remove_inventory_folder = self._wrap_with_cache_defer(
            self._handle_remove_inventory_folder
        )
        self._handle_bulk_update_inventory = self._wrap_with_cache_defer(
            self._handle_bulk_update_inventory
        )
        self._handle_move_inventory_item = self._wrap_with_cache_defer(
            self._handle_move_inventory_item
        )
        self._handle_move_inventory_folder = self._wrap_with_cache_defer(
            self._handle_move_inventory_folder
        )
        self.process_aisv3_response = self._wrap_with_cache_defer(
            self.process_aisv3_response
        )

        # Base constructor after, because it registers handlers to specific methods, which need to
        # be wrapped before we call they're registered. Handlers are registered by method reference,
        # not by name!
        super().__init__(session)
        session.http_message_handler.subscribe("InventoryAPIv3", self._handle_aisv3_flow)
        newest_cache = None
        newest_timestamp = dt.datetime(year=1970, month=1, day=1, tzinfo=dt.timezone.utc)
        # So consumers know when the inventory should be complete
        self.cache_loaded: asyncio.Event = asyncio.Event()
        self._cache_deferred_calls: List[Tuple[Callable[..., None], Tuple]] = []
        # Look for the newest version of the cached inventory and use that.
        # Not foolproof, but close enough if we're not sure what viewer is being used.
        for cache_dir in iter_viewer_cache_dirs():
            inv_cache_path = cache_dir / (str(session.agent_id) + ".inv.llsd.gz")
            if inv_cache_path.exists():
                mod = get_mtime(inv_cache_path)
                if not mod:
                    continue
                mod_ts = dt.datetime.fromtimestamp(mod, dt.timezone.utc)
                if mod_ts <= newest_timestamp:
                    continue
                newest_cache = inv_cache_path

        if newest_cache:
            cache_load_fut = asyncio.ensure_future(asyncio.to_thread(self.load_cache, newest_cache))
            # Meh. Don't care if it fails.
            cache_load_fut.add_done_callback(lambda *args: self.cache_loaded.set())
            create_logged_task(self._apply_deferred_after_loaded(), "Apply deferred inventory", LOG)
        else:
            self.cache_loaded.set()

    async def _apply_deferred_after_loaded(self):
        await self.cache_loaded.wait()
        LOG.info("Applying deferred inventory calls")
        deferred_calls = self._cache_deferred_calls[:]
        self._cache_deferred_calls.clear()
        for func, args in deferred_calls:
            try:
                func(*args)
            except:
                LOG.exception("Failed to apply deferred inventory call")

    def _wrap_with_cache_defer(self, func: Callable[..., None]):
        @functools.wraps(func)
        def wrapped(*inner_args):
            if not self.cache_loaded.is_set():
                self._cache_deferred_calls.append((func, inner_args))
            else:
                func(*inner_args)
        return wrapped

    def _handle_aisv3_flow(self, flow: HippoHTTPFlow):
        if flow.response.status_code < 200 or flow.response.status_code > 300:
            # Probably not a success
            return
        content_type = flow.response.headers.get("Content-Type", "")
        if "llsd" not in content_type:
            # Okay, probably still some kind of error...
            return

        # Try and add anything from the response into the model
        self.process_aisv3_response(llsd.parse(flow.response.content))

    async def create_folder(
            self,
            parent: InventoryCategory | UUID,
            name: str,
            pref_type: int = AssetType.NONE,
            cat_id: UUID | None = None
    ) -> InventoryCategory:
        cat = await super().create_folder(parent, name, pref_type, cat_id)
        await self._session.main_region.circuit.send_reliable(self._craft_update_message(cat))
        return cat

    async def create_item(
            self,
            parent: UUID | InventoryCategory,
            name: str,
            type: AssetType,
            inv_type: InventoryType,
            wearable_type: WearableType,
            transaction_id: UUID,
            next_mask: int | Permissions = 0x0008e000,
            description: str = '',
    ) -> InventoryItem:
        item = await super().create_item(
            parent=parent,
            name=name,
            type=type,
            inv_type=inv_type,
            wearable_type=wearable_type,
            transaction_id=transaction_id,
            next_mask=next_mask,
            description=description,
        )
        await self._session.main_region.circuit.send_reliable(self._craft_update_message(item))
        return item

    async def update(self, node: InventoryNodeBase, data: dict) -> None:
        await super().update(node, data)
        await self._session.main_region.circuit.send_reliable(self._craft_update_message(node))

    async def move(self, node: InventoryNodeBase, new_parent: UUID | InventoryCategory) -> None:
        await super().move(node, new_parent)
        await self._session.main_region.circuit.send_reliable(self._craft_update_message(node))

    async def copy(self, node: InventoryNodeBase, destination: UUID | InventoryCategory, contents: bool = True)\
            -> InventoryCategory | InventoryItem:
        ret_node = await super().copy(node, destination, contents)
        await self._session.main_region.circuit.send_reliable(self._craft_update_message(node))
        return ret_node

    def _craft_removal_message(self, node: InventoryNodeBase) -> Message:
        is_folder = True
        if isinstance(node, InventoryItem):
            is_folder = False

        msg = Message(
            "RemoveInventoryFolder" if is_folder else "RemoveInventoryItem",
            Block("AgentData", AgentID=self._session.agent_id, SessionID=self._session.id),
            direction=Direction.IN,
        )
        if is_folder:
            msg.add_block(Block("FolderData", FolderID=node.node_id))
        else:
            msg.add_block(Block("InventoryData", ItemID=node.node_id))
        return msg

    def _craft_update_message(self, node: InventoryNodeBase):
        msg = Message(
            "BulkUpdateInventory",
            Block("AgentData", AgentID=self._session.agent_id, TransactionID=UUID.random()),
            direction=Direction.IN,
        )

        if isinstance(node, InventoryItem):
            msg.add_block(node.to_inventory_data("ItemData"))
        elif isinstance(node, InventoryCategory):
            msg.add_block(node.to_folder_data())
        return msg



import asyncio
import math
import random
import unittest
from typing import *
from unittest import mock

from hippolyzer.lib.base.datatypes import *
from hippolyzer.lib.base.helpers import create_logged_task
from hippolyzer.lib.base.message.message import Block, Message as Message
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer
from hippolyzer.lib.base.objects import Object, normalize_object_update_compressed_data
from hippolyzer.lib.base.templates import ExtraParamType, PCode, JUST_CREATED_FLAGS
from hippolyzer.lib.client.object_manager import ObjectUpdateType
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.vocache import RegionViewerObjectCacheChain, RegionViewerObjectCache, ViewerObjectCacheEntry
from hippolyzer.lib.proxy.test_utils import BaseProxyTest

OBJECT_UPDATE_COMPRESSED_DATA = (
    b"\x12\x12\x10\xbf\x16XB~\x8f\xb4\xfb\x00\x1a\xcd\x9b\xe5\xd2\x04\x00\x00\t\x00\xcdG\x00\x00"
    b"\x03\x00\x00\x00\x1cB\x00\x00\x1cB\xcd\xcc\xcc=\xedG,"
    b"B\x9e\xb1\x9eBff\xa0A\x00\x00\x00\x00\x00\x00\x00\x00["
    b"\x8b\xf8\xbe\xc0\x00\x00\x00k\x9b\xc4\xfe3\nOa\xbb\xe2\xe4\xb2C\xac7\xbd\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\xa2=\x010\x00\x11\x00\x00\x00\x89UgG$\xcbC\xed\x92\x0bG\xca\xed"
    b"\x15F_@ \x00\x00\x00\x00d\x96\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00?\x00\x00\x00\x1c\x9fJoI\x8dH\xa0\x9d\xc4&''\x19=g\x00\x00\x00\x003\x00ff\x86\xbf"
    b"\x00ff\x86?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89UgG$\xcbC"
    b"\xed\x92\x0bG\xca\xed\x15F_\x10\x00\x00\x003\x00\x01\x01\x00\x00\x00\x00\xdb\x0f\xc9@\xa6"
    b"\x9b\xc4="
)


class WrappingMessageHandler:
    """Calls both the session and region-local message handlers"""
    def __init__(self, region: ProxiedRegion):
        self.region = region

    def handle(self, message: Message):
        message.sender = self.region.circuit_addr
        self.region.session().message_handler.handle(message)
        self.region.message_handler.handle(message)


class ObjectTrackingAddon(BaseAddon):
    def __init__(self):
        super().__init__()
        self.events = []

    def handle_object_updated(self, session, region, obj: Object, updated_props: Set[str], msg: Optional[Message]):
        self.events.append(("update", obj, updated_props))

    def handle_object_killed(self, session, region, obj: Object):
        self.events.append(("kill", obj))


class ObjectManagerTestMixin(BaseProxyTest):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self._setup_default_circuit()
        self.region = self.session.main_region
        self.message_handler = WrappingMessageHandler(self.region)
        patched = mock.patch('hippolyzer.lib.proxy.vocache.RegionViewerObjectCacheChain.for_region')
        self.addCleanup(patched.stop)
        self.mock_get_region_object_cache_chain = patched.start()
        self.mock_get_region_object_cache_chain.return_value = RegionViewerObjectCacheChain([])
        self.region_object_manager = self.region.objects
        self.serializer = UDPMessageSerializer()
        self.deserializer = UDPMessageDeserializer()
        self.object_addon = ObjectTrackingAddon()
        AddonManager.init([], None, [self.object_addon])

    def _create_object_update(self, local_id=None, full_id=None, parent_id=None, pos=None, rot=None,
                              pcode=None, namevalue=None, region_handle=None) -> Message:
        pos = pos if pos is not None else (1.0, 2.0, 3.0)
        rot = rot if rot is not None else (0.0, 0.0, 0.0, 1.0)
        pcode = pcode if pcode is not None else PCode.PRIMITIVE
        if region_handle is None:
            region_handle = 123
        msg = Message(
            "ObjectUpdate",
            Block("RegionData", RegionHandle=region_handle, TimeDilation=123),
            Block(
                "ObjectData",
                ID=local_id if local_id is not None else random.getrandbits(32),
                FullID=full_id if full_id else UUID.random(),
                PCode=pcode,
                Scale=Vector3(0.5, 0.5, 0.5),
                UpdateFlags=268568894,
                PathCurve=16,
                ParentID=parent_id if parent_id else 0,
                ProfileCurve=1,
                PathScaleX=100,
                PathScaleY=100,
                NameValue=namevalue,
                TextureEntry=b'\x89UgG$\xcbC\xed\x92\x0bG\xca\xed\x15F_\x00\x00\x00\x00\x00\x00\x00\x00\x80?\x00\x00'
                             b'\x00\x80?\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                             b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
                TextColor=b'\x00\x00\x00\x00',
                ExtraParams=b'\x00',
                fill_missing=True,
            )
        )
        msg["ObjectData"][0].serialize_var(
            "ObjectData",
            (
                60,
                {
                    'Position': pos,
                    'Velocity': (0.0, 0.0, 0.0),
                    'Acceleration': (0.0, 0.0, 0.0),
                    'Rotation': rot,
                    'AngularVelocity': (0.0, 0.0, 0.0)
                }
            )
        )
        # Run through (de)serializer to fill in any missing vars
        return self.deserializer.deserialize(self.serializer.serialize(msg))

    def _create_object(self, local_id=None, full_id=None, parent_id=None, pos=None, rot=None,
                       pcode=None, namevalue=None, region_handle=None) -> Object:
        msg = self._create_object_update(
            local_id=local_id, full_id=full_id, parent_id=parent_id, pos=pos, rot=rot,
            pcode=pcode, namevalue=namevalue, region_handle=region_handle)
        self.message_handler.handle(msg)
        actual_handle = msg["RegionData"]["RegionHandle"]
        region = self.session.region_by_handle(actual_handle)
        return region.objects.lookup_fullid(msg["ObjectData"]["FullID"])

    def _create_kill_object(self, local_id) -> Message:
        return Message(
            "KillObject",
            Block(
                "ObjectData",
                ID=local_id,
            )
        )

    def _kill_object(self, local_id: int):
        self.message_handler.handle(self._create_kill_object(local_id))

    def _create_object_update_cached(self, local_id: int, region_handle: int = 123,
                                     crc: int = 22, flags: int = 4321):
        return Message(
            'ObjectUpdateCached',
            Block("RegionData", TimeDilation=102, RegionHandle=region_handle),
            Block(
                "ObjectData",
                ID=local_id,
                CRC=crc,
                UpdateFlags=flags,
            )
        )

    def _get_avatar_positions(self) -> Dict[UUID, Vector3]:
        return {av.FullID: av.RegionPosition for av in self.region_object_manager.all_avatars}


class RegionObjectManagerTests(ObjectManagerTestMixin, unittest.IsolatedAsyncioTestCase):
    def test_basic_tracking(self):
        """Does creating an object result in it being tracked?"""
        msg = self._create_object_update()
        self.message_handler.handle(msg)
        obj = self.region_object_manager.lookup_fullid(msg["ObjectData"]["FullID"])
        self.assertIsNotNone(obj)

    def test_terse_object_update(self):
        msg = self._create_object_update(pos=Vector3(1, 2, 3))
        self.message_handler.handle(msg)
        local_id = msg["ObjectData"]["ID"]
        msg = Message(
            'ImprovedTerseObjectUpdate',
            Block('RegionData', RegionHandle=123, TimeDilation=65345),
            Block(
                'ObjectData',
                Data_={
                    'ID': local_id,
                    'State': 0,
                    'FootCollisionPlane': None,
                    'Position': Vector3(-2, -3, -4),
                    'Velocity': Vector3(-0.0, -0.0, -0.0),
                    'Acceleration': Vector3(-0.0, -0.0, -0.0),
                    'Rotation': Quaternion(0, 0, 0, 1),
                    'AngularVelocity': Vector3(-0.0, -0.0, -0.0)
                },
                TextureEntry_=None,
            ),
        )
        self.message_handler.handle(msg)
        obj = self.region_object_manager.lookup_localid(local_id)
        self.assertEqual(obj.Position, Vector3(-2, -3, -4))

    def test_parent_tracking(self):
        """Are basic parenting scenarios handled?"""
        parent = self._create_object()
        child = self._create_object(parent_id=parent.LocalID)
        self.assertSequenceEqual([child.LocalID], parent.ChildIDs)

    def test_orphan_parent_tracking(self):
        child = self._create_object(local_id=2, parent_id=1)
        self.assertEqual({1}, self.region_object_manager.missing_locals)
        parent = self._create_object(local_id=1)
        self.assertEqual(set(), self.region_object_manager.missing_locals)
        self.assertSequenceEqual([child.LocalID], parent.ChildIDs)

    def test_killing_parent_kills_children(self):
        _child = self._create_object(local_id=2, parent_id=1)
        parent = self._create_object(local_id=1)
        # This should orphan the child again
        self._kill_object(parent.LocalID)
        parent = self._create_object(local_id=1)
        # We should not have picked up any children
        self.assertSequenceEqual([], parent.ChildIDs)

    def test_hierarchy_killed(self):
        _child = self._create_object(local_id=3, parent_id=2)
        _other_child = self._create_object(local_id=4, parent_id=2)
        _parent = self._create_object(local_id=2, parent_id=1)
        grandparent = self._create_object(local_id=1)
        # KillObject implicitly kills all known descendents at that point
        self._kill_object(grandparent.LocalID)
        self.assertEqual(0, len(self.region_object_manager))

    def test_hierarchy_avatar_not_killed(self):
        _child = self._create_object(local_id=3, parent_id=2)
        _parent = self._create_object(local_id=2, parent_id=1, pcode=PCode.AVATAR)
        grandparent = self._create_object(local_id=1)
        # KillObject should only "unsit" child avatars (does this require an ObjectUpdate
        # or is ParentID=0 implied?)
        self._kill_object(grandparent.LocalID)
        self.assertEqual(2, len(self.region_object_manager))
        self.assertIsNotNone(self.region_object_manager.lookup_localid(2))

    def test_attachment_orphan_parent_tracking(self):
        """
        Test that multi-level parenting trees handle orphaning correctly.

        Technically there can be at least 4 levels of parenting if sitting.
        object -> seated agent -> attachment root -> attachment child
        """
        child = self._create_object(local_id=3, parent_id=2)
        parent = self._create_object(local_id=2, parent_id=1)
        self.assertSequenceEqual([child.LocalID], parent.ChildIDs)

    def test_unparenting_succeeds(self):
        child = self._create_object(local_id=3, parent_id=2)
        parent = self._create_object(local_id=2)
        msg = self._create_object_update(local_id=child.LocalID, full_id=child.FullID, parent_id=0)
        self.message_handler.handle(msg)
        self.assertEqual(0, child.ParentID)
        self.assertSequenceEqual([], parent.ChildIDs)

    def test_reparenting_succeeds(self):
        child = self._create_object(local_id=3, parent_id=2)
        parent = self._create_object(local_id=2)
        second_parent = self._create_object(local_id=1)
        msg = self._create_object_update(local_id=child.LocalID,
                                         full_id=child.FullID, parent_id=second_parent.LocalID)
        self.message_handler.handle(msg)
        self.assertEqual(second_parent.LocalID, child.ParentID)
        self.assertSequenceEqual([], parent.ChildIDs)
        self.assertSequenceEqual([child.LocalID], second_parent.ChildIDs)

    def test_reparenting_without_known_parent_succeeds(self):
        child = self._create_object(local_id=3, parent_id=2)
        second_parent = self._create_object(local_id=1)
        msg = self._create_object_update(local_id=child.LocalID,
                                         full_id=child.FullID, parent_id=second_parent.LocalID)
        self.message_handler.handle(msg)
        # Create the original parent after its former child has been reparented
        parent = self._create_object(local_id=2)
        self.assertEqual(second_parent.LocalID, child.ParentID)
        self.assertSequenceEqual([], parent.ChildIDs)
        self.assertSequenceEqual([child.LocalID], second_parent.ChildIDs)

    def test_reparenting_with_neither_parent_known_succeeds(self):
        child = self._create_object(local_id=3, parent_id=2)
        msg = self._create_object_update(local_id=child.LocalID,
                                         full_id=child.FullID, parent_id=1)
        self.message_handler.handle(msg)
        second_parent = self._create_object(local_id=1)
        self.assertEqual(second_parent.LocalID, child.ParentID)
        self.assertSequenceEqual([child.LocalID], second_parent.ChildIDs)

    def test_property_changes_reported_correctly(self):
        obj = self._create_object(local_id=1)
        msg = self._create_object_update(local_id=obj.LocalID, full_id=obj.FullID, pos=(2.0, 2.0, 2.0))
        self.message_handler.handle(msg)
        events = self.object_addon.events
        self.assertEqual(2, len(events))
        self.assertEqual({"Position", "TextureEntry"}, events[1][2])

    def test_region_position(self):
        parent = self._create_object(pos=(0.0, 1.0, 0.0))
        child = self._create_object(parent_id=parent.LocalID, pos=(0.0, 1, 0.0))

        self.assertEqual(parent.RegionPosition, (0.0, 1.0, 0.0))
        self.assertEqual(child.RegionPosition, (0.0, 2.0, 0.0))

    def test_orphan_region_position(self):
        child = self._create_object(local_id=2, parent_id=1, pos=(0.0, 1, 0.0))

        with self.assertRaises(ValueError):
            getattr(child, "RegionPosition")

    def test_rotated_region_position(self):
        parent = self._create_object(pos=(0.0, 1.0, 0.0), rot=Quaternion.from_euler(0, 0, 180, True))
        child = self._create_object(parent_id=parent.LocalID, pos=(0.0, 1.0, 0.0))

        self.assertEqual(parent.RegionPosition, (0.0, 1.0, 0.0))
        self.assertEqual(child.RegionPosition, (0.0, 0.0, 0.0))

    def test_rotated_region_position_multi_level(self):
        rot = Quaternion.from_euler(0, 0, 180, True)
        grandparent = self._create_object(pos=(0.0, 1.0, 0.0), rot=rot)
        parent = self._create_object(parent_id=grandparent.LocalID, pos=(0.0, 1.0, 0.0), rot=rot)
        child = self._create_object(parent_id=parent.LocalID, pos=(1.0, 2.0, 0.0))

        self.assertEqual(grandparent.RegionPosition, (0.0, 1.0, 0.0))
        self.assertEqual(parent.RegionPosition, (0.0, 0.0, 0.0))
        self.assertEqual(child.RegionPosition, (1.0, 2.0, 0.0))

    def test_global_position(self):
        obj = self._create_object(pos=(0.0, 0.0, 0.0))
        self.assertEqual(obj.GlobalPosition, (0.0, 123.0, 0.0))

    def test_avatar_locations(self):
        agent1_id = UUID.random()
        agent2_id = UUID.random()
        self.message_handler.handle(Message(
            "CoarseLocationUpdate",
            Block("AgentData", AgentID=agent1_id),
            Block("AgentData", AgentID=agent2_id),
            Block("Location", X=1, Y=2, Z=3),
            Block("Location", X=2, Y=3, Z=4),
        ))
        self.assertDictEqual(self._get_avatar_positions(), {
            # CoarseLocation's Z axis is multiplied by 4
            agent1_id: Vector3(1, 2, 12),
            agent2_id: Vector3(2, 3, 16),
        })

        # Simulate an avatar sitting on an object
        seat_object = self._create_object(pos=(0, 0, 3))
        # If we have a real object pos it should override coarse pos
        avatar_obj = self._create_object(full_id=agent1_id, pcode=PCode.AVATAR,
                                         parent_id=seat_object.LocalID, pos=Vector3(0, 0, 2))
        self.assertDictEqual(self._get_avatar_positions(), {
            # Agent is seated, make sure this is region and not local pos
            agent1_id: Vector3(0, 0, 5),
            agent2_id: Vector3(2, 3, 16),
        })

        # Simulate missing parent for agent
        self._kill_object(seat_object.LocalID)
        self.assertDictEqual(self._get_avatar_positions(), {
            # Agent is seated, but we don't know its parent. We have
            # to use the coarse location.
            agent1_id: Vector3(1, 2, 12),
            agent2_id: Vector3(2, 3, 16),
        })

        # If the object is killed and no coarse pos, it shouldn't be in the dict
        # CoarseLocationUpdates are expected to be complete, so any agents missing
        # are no longer in the sim.
        self._kill_object(avatar_obj.LocalID)
        self.message_handler.handle(Message(
            "CoarseLocationUpdate",
            Block("AgentData", AgentID=agent2_id),
            Block("Location", X=2, Y=3, Z=4),
        ))
        self.assertDictEqual(self._get_avatar_positions(), {
            agent2_id: Vector3(2, 3, 16),
        })

        # 255 on Z axis means we can't guess the real Z
        self.message_handler.handle(Message(
            "CoarseLocationUpdate",
            Block("AgentData", AgentID=agent2_id),
            Block("Location", X=2, Y=3, Z=math.inf),
        ))
        self.assertDictEqual(self._get_avatar_positions(), {
            agent2_id: Vector3(2, 3, math.inf),
        })
        agent2_avatar = self.region_object_manager.lookup_avatar(agent2_id)
        self.assertEqual(agent2_avatar.GlobalPosition, Vector3(2, 126, math.inf))

    def test_name_cache(self):
        # Receiving an update with a NameValue for an avatar should update NameCache
        obj = self._create_object(
            pcode=PCode.AVATAR,
            namevalue='DisplayName STRING RW DS 𝔲𝔫𝔦𝔠𝔬𝔡𝔢𝔫𝔞𝔪𝔢\n'
                      'FirstName STRING RW DS firstname\n'
                      'LastName STRING RW DS Resident\n'
                      'Title STRING RW DS foo'.encode("utf8"),
        )
        self.assertEqual(self.session_manager.name_cache.lookup(obj.FullID).first_name, "firstname")
        av = self.region_object_manager.lookup_avatar(obj.FullID)
        self.assertEqual(av.Name, "𝔲𝔫𝔦𝔠𝔬𝔡𝔢𝔫𝔞𝔪𝔢 (firstname Resident)")
        self.assertEqual(av.PreferredName, "𝔲𝔫𝔦𝔠𝔬𝔡𝔢𝔫𝔞𝔪𝔢")

    def test_normalize_cache_data(self):
        normalized = normalize_object_update_compressed_data(OBJECT_UPDATE_COMPRESSED_DATA)
        expected = {
            'PSBlock': None,
            'ParentID': 0,
            'LocalID': 1234,
            'FullID': UUID('121210bf-1658-427e-8fb4-fb001acd9be5'),
            'PCode': PCode.PRIMITIVE,
            'State': 0,
            'CRC': 18381,
            'Material': 3,
            'ClickAction': 0,
            'Scale': Vector3(39.0, 39.0, 0.10000000149011612),
            'Position': Vector3(43.07024002075195, 79.34690856933594, 20.049999237060547),
            'Rotation': Quaternion(0.0, 0.0, -0.48543819785118103, 0.8742709854884798),
            'OwnerID': UUID('6b9bc4fe-330a-4f61-bbe2-e4b243ac37bd'),
            'AngularVelocity': Vector3(0.0, 0.0, 0.0791015625),
            'TreeSpecies': None,
            'ScratchPad': None,
            'Text': b'',
            'TextColor': b'',
            'MediaURL': b'',
            'Sound': UUID(),
            'SoundGain': 0.0,
            'SoundFlags': 0,
            'SoundRadius': 0.0,
            'NameValue': [],
            'PathCurve': 32,
            'ProfileCurve': 0,
            'PathBegin': 0,
            'PathEnd': 25600,
            'PathScaleX': 150,
            'PathScaleY': 0,
            'PathShearX': 0,
            'PathShearY': 0,
            'PathTwist': 0,
            'PathTwistBegin': 0,
            'PathRadiusOffset': 0,
            'PathTaperX': 0,
            'PathTaperY': 0,
            'PathRevolutions': 0,
            'PathSkew': 0,
            'ProfileBegin': 0,
            'ProfileEnd': 0,
            'ProfileHollow': 0,
        }
        filtered_normalized = {k: v for k, v in normalized.items() if k in expected}
        self.assertDictEqual(filtered_normalized, expected)
        sculpt_texture = normalized["ExtraParams"][ExtraParamType.SCULPT]["Texture"]
        self.assertEqual(sculpt_texture, UUID('89556747-24cb-43ed-920b-47caed15465f'))
        self.assertIsNotNone(normalized['TextureAnim'])
        self.assertIsNotNone(normalized['TextureEntry'])

    def test_object_cache(self):
        self.mock_get_region_object_cache_chain.return_value = RegionViewerObjectCacheChain([
            RegionViewerObjectCache(self.region.cache_id, [
                ViewerObjectCacheEntry(
                    local_id=1234,
                    crc=22,
                    data=OBJECT_UPDATE_COMPRESSED_DATA,
                )
            ])
        ])
        cache_msg = self._create_object_update_cached(1234, flags=4321)
        obj = self.region_object_manager.lookup_localid(1234)
        self.assertIsNone(obj)
        self.region_object_manager.load_cache()
        self.message_handler.handle(cache_msg)
        obj = self.region_object_manager.lookup_localid(1234)
        self.assertEqual(obj.FullID, UUID('121210bf-1658-427e-8fb4-fb001acd9be5'))
        # Flags from the ObjectUpdateCached should have been merged in
        self.assertEqual(obj.UpdateFlags, 4321)

    async def test_request_objects(self):
        # request five objects, three of which won't receive an ObjectUpdate
        futures = self.region_object_manager.request_objects((1234, 1235, 1236, 1237))
        self._create_object(1234)
        self._create_object(1235)
        done, pending = await asyncio.wait(futures, timeout=0.0001)
        objects = await asyncio.gather(*done)
        # wait() returns unordered results, so use a set.
        self.assertEqual(set(o.LocalID for o in objects), {1234, 1235})
        pending = list(pending)
        self.assertEqual(2, len(pending))
        pending_1, pending_2 = pending

        # Timing out should cancel
        with self.assertRaises(asyncio.TimeoutError):
            await asyncio.wait_for(pending_1, 0.00001)
        self.assertTrue(pending_1.cancelled())

        fut = self.region_object_manager.request_objects(1238)[0]
        self._kill_object(1238)
        self.assertTrue(fut.cancelled())

        # Object manager being cleared due to region death should cancel
        self.assertFalse(pending_2.cancelled())
        self.region_object_manager.clear()
        self.assertTrue(pending_2.cancelled())
        # The clear should have triggered the objects to be removed from the world view as well
        # as the region view
        self.assertEqual(0, len(self.session.objects))
        self.assertEqual(0, len(self.region_object_manager))


class SessionObjectManagerTests(ObjectManagerTestMixin, unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        self.second_region = self.session.register_region(
            ("127.0.0.1", 9), "https://localhost:5", 124
        )
        self.session.objects.track_region_objects(124)
        self._setup_region_circuit(self.second_region)

    def test_get_fullid(self):
        obj = self._create_object()
        self.assertIs(self.session.objects.lookup_fullid(obj.FullID), obj)
        self._kill_object(obj.LocalID)
        self.assertIsNone(self.session.objects.lookup_fullid(obj.FullID))

    def test_region_handle_change(self):
        obj = self._create_object(region_handle=123)
        self.assertEqual(obj.RegionHandle, 123)
        self.assertIs(self.region.objects.lookup_fullid(obj.FullID), obj)
        self.assertIs(self.region.objects.lookup_localid(obj.LocalID), obj)

        # Send an update moving the object to the new region
        self._create_object(local_id=~obj.LocalID & 0xFFffFFff, full_id=obj.FullID, region_handle=124)
        self.assertEqual(obj.RegionHandle, 124)
        self.assertIsNone(self.region.objects.lookup_fullid(obj.FullID))
        self.assertIsNone(self.region.objects.lookup_localid(obj.LocalID))
        self.assertIs(self.second_region.objects.lookup_fullid(obj.FullID), obj)
        self.assertIs(self.second_region.objects.lookup_localid(obj.LocalID), obj)
        self.assertEqual(1, len(self.session.objects))
        self.assertEqual(0, len(self.region.objects))
        self.assertEqual(1, len(self.second_region.objects))

    def test_object_moved_to_bad_region(self):
        obj = self._create_object(region_handle=123)
        msg = self._create_object_update(
            local_id=~obj.LocalID & 0xFFffFFff, full_id=obj.FullID, region_handle=999)
        self.message_handler.handle(msg)
        # Should not be in this region anymore
        self.assertEqual(0, len(self.region_object_manager))
        # Should still be tracked by the session
        self.assertEqual(1, len(self.session.objects))
        self.assertIsNotNone(self.session.objects.lookup_fullid(obj.FullID))

    def test_linkset_region_handle_change(self):
        parent = self._create_object(region_handle=123)
        child = self._create_object(region_handle=123, parent_id=parent.LocalID)
        self._create_object(local_id=~parent.LocalID & 0xFFffFFff, full_id=parent.FullID, region_handle=124)
        # Children reference their parents, not the other way around. Moving this to a new region
        # should have cleared the list because it now has no children in the same region.
        self.assertEqual([], parent.ChildIDs)
        # Move the child to the same region
        self._create_object(
            local_id=child.LocalID, full_id=child.FullID, region_handle=124, parent_id=parent.LocalID)
        # Child should be back in the children list
        self.assertEqual([child.LocalID], parent.ChildIDs)
        self.assertEqual(parent.LocalID, child.ParentID)
        self.assertEqual(0, len(self.region.objects))
        self.assertEqual(2, len(self.second_region.objects))
        self.assertEqual(0, len(self.region.objects.missing_locals))
        self.assertEqual(0, len(self.second_region.objects.missing_locals))

    def test_all_objects(self):
        obj = self._create_object()
        self.assertEqual([obj], list(self.session.objects.all_objects))

    def test_all_avatars(self):
        obj = self._create_object(pcode=PCode.AVATAR)
        av_list = list(self.session.objects.all_avatars)
        self.assertEqual(1, len(av_list))
        self.assertEqual(obj, av_list[0].Object)

    def test_avatars_preference(self):
        # If we have a coarselocation for an avatar in one region and
        # an actual object in another, we should always prefer the
        # one with the actual object.
        av_1 = self._create_object(pcode=PCode.AVATAR, region_handle=123)
        av_2 = self._create_object(pcode=PCode.AVATAR, region_handle=124)
        # Coarse location shouldn't be used for either of these
        self.session.region_by_handle(123).message_handler.handle(Message(
            "CoarseLocationUpdate",
            Block("AgentData", AgentID=av_2.FullID),
            Block("Location", X=2, Y=3, Z=4),
        ))
        self.session.region_by_handle(124).message_handler.handle(Message(
            "CoarseLocationUpdate",
            Block("AgentData", AgentID=av_1.FullID),
            Block("Location", X=2, Y=3, Z=4),
        ))
        av_list = list(self.session.objects.all_avatars)
        self.assertEqual(2, len(av_list))
        self.assertTrue(all(a.Object for a in av_list))

    def test_lookup_avatar(self):
        av_1 = self._create_object(pcode=PCode.AVATAR)
        av_obj = self.session.objects.lookup_avatar(av_1.FullID)
        self.assertEqual(av_1.FullID, av_obj.FullID)

    async def test_requesting_properties(self):
        obj = self._create_object()
        futs = self.session.objects.request_object_properties(obj)
        self.message_handler.handle(Message(
            "ObjectProperties",
            Block("ObjectData", ObjectID=obj.FullID, Name="Foobar", TextureID=b""),
        ))
        await asyncio.wait_for(futs[0], timeout=0.0001)
        self.assertEqual(obj.Name, "Foobar")

    async def test_load_ancestors(self):
        child = self._create_object(region_handle=123, parent_id=1)
        parentless = self._create_object(region_handle=123)
        orphaned = self._create_object(region_handle=123, parent_id=9)

        async def _create_after():
            await asyncio.sleep(0.001)
            self._create_object(region_handle=123, local_id=child.ParentID)
        _ = create_logged_task(_create_after())

        await self.session.objects.load_ancestors(child)
        await self.session.objects.load_ancestors(parentless)
        with self.assertRaises(asyncio.TimeoutError):
            await self.session.objects.load_ancestors(orphaned, wait_time=0.005)

    async def test_auto_request_objects(self):
        self.session_manager.settings.AUTOMATICALLY_REQUEST_MISSING_OBJECTS = True
        self.message_handler.handle(self._create_object_update_cached(1234))
        self.message_handler.handle(self._create_object_update_cached(1235))
        self.assertEqual({1234, 1235}, self.region_object_manager.queued_cache_misses)
        # Pretend viewer sent out its own RequestMultipleObjects
        self.message_handler.handle(Message(
            'RequestMultipleObjects',
            Block("RegionData", SessionID=self.session.id, AgentID=self.session.agent_id),
            Block(
                "ObjectData",
                ID=1234,
            )
        ))
        # Proxy should have killed its pending request for 1234
        self.assertEqual({1235}, self.region_object_manager.queued_cache_misses)

    async def test_auto_request_avatar_seats(self):
        # Avatars' parent links should always be requested regardless of
        # object auto-request setting's value.
        seat_id = 999
        av = self._create_object(pcode=PCode.AVATAR, parent_id=seat_id)
        self.assertEqual({seat_id}, self.region_object_manager.queued_cache_misses)
        # Need to wait for it to decide it's worth requesting
        await asyncio.sleep(0.22)
        self.assertEqual(set(), self.region_object_manager.queued_cache_misses)
        # Make sure we sent a request after the timeout
        req_msg = self.deserializer.deserialize(self.transport.packets[-1][0])
        self.assertEqual("RequestMultipleObjects", req_msg.name)
        self.assertEqual(
            [{'CacheMissType': 0, 'ID': seat_id}],
            req_msg.to_dict()['body']['ObjectData'],
        )
        # Parent should not be requested again if an unrelated property like pos changes
        self._create_object(local_id=av.LocalID, full_id=av.FullID,
                            pcode=PCode.AVATAR, parent_id=seat_id, pos=(1, 2, 9))
        self.assertEqual(set(), self.region_object_manager.queued_cache_misses)

    async def test_handle_object_update_event(self):
        with self.session.objects.events.subscribe_async(
            message_names=(ObjectUpdateType.UPDATE,),
            predicate=lambda e: e.object.UpdateFlags & JUST_CREATED_FLAGS and "LocalID" in e.updated,
        ) as get_events:
            self._create_object(local_id=999)
            evt = await asyncio.wait_for(get_events(), 1.0)
            self.assertEqual(999, evt.object.LocalID)

    async def test_handle_object_update_predicate(self):
        with self.session.objects.events.subscribe_async(
            message_names=(ObjectUpdateType.UPDATE,),
        ) as get_events:
            self._create_object(local_id=999)
            evt = await asyncio.wait_for(get_events(), 1.0)
            self.assertEqual(999, evt.object.LocalID)

    async def test_handle_object_update_events_two_subscribers(self):
        with self.session.objects.events.subscribe_async(
            message_names=(ObjectUpdateType.UPDATE,),
        ) as get_events:
            with self.session.objects.events.subscribe_async(
                    message_names=(ObjectUpdateType.UPDATE,),
            ) as get_events2:
                self._create_object(local_id=999)
                evt = await asyncio.wait_for(get_events(), 1.0)
                evt2 = await asyncio.wait_for(get_events2(), 1.0)
                self.assertEqual(999, evt.object.LocalID)
                self.assertEqual(evt, evt2)

    async def test_handle_object_update_events_two_subscribers_timeout(self):
        with self.session.objects.events.subscribe_async(
            message_names=(ObjectUpdateType.UPDATE,),
        ) as get_events:
            with self.session.objects.events.subscribe_async(
                    message_names=(ObjectUpdateType.UPDATE,),
            ) as get_events2:
                self._create_object(local_id=999)
                evt = asyncio.wait_for(get_events(), 0.01)
                evt2 = asyncio.wait_for(get_events2(), 0.01)
                await asyncio.gather(evt, evt2)



from __future__ import annotations

import asyncio
import hashlib
from importlib.metadata import version
import logging
import uuid
import weakref
import xmlrpc.client
from typing import *

import aiohttp
import multidict

from hippolyzer.lib.base.datatypes import Vector3, StringEnum
from hippolyzer.lib.base.helpers import proxify, get_resource_filename, create_logged_task
from hippolyzer.lib.base.message.circuit import Circuit
from hippolyzer.lib.base.message.llsd_msg_serializer import LLSDMessageSerializer
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.message.message_dot_xml import MessageDotXML
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.network.caps_client import CapsClient, CAPS_DICT
from hippolyzer.lib.base.network.transport import ADDR_TUPLE, Direction, SocketUDPTransport, AbstractUDPTransport
from hippolyzer.lib.base.settings import Settings, SettingDescriptor
from hippolyzer.lib.base.templates import RegionHandshakeReplyFlags, ChatType, ThrottleData
from hippolyzer.lib.base.transfer_manager import TransferManager
from hippolyzer.lib.base.xfer_manager import XferManager
from hippolyzer.lib.client.asset_uploader import AssetUploader
from hippolyzer.lib.client.inventory_manager import InventoryManager
from hippolyzer.lib.client.object_manager import ClientObjectManager, ClientWorldObjectManager
from hippolyzer.lib.client.parcel_manager import ParcelManager
from hippolyzer.lib.client.state import BaseClientSession, BaseClientRegion, BaseClientSessionManager


LOG = logging.getLogger(__name__)


class StartLocation(StringEnum):
    LAST = "last"
    HOME = "home"


class ClientSettings(Settings):
    SSL_VERIFY: bool = SettingDescriptor(False)
    """Off by default for now, the cert validation is a big mess due to LL using an internal CA."""
    SSL_CERT_PATH: str = SettingDescriptor(get_resource_filename("lib/base/network/data/ca-bundle.crt"))
    USER_AGENT: str = SettingDescriptor(f"Hippolyzer/v{version('hippolyzer')}")
    SEND_AGENT_UPDATES: bool = SettingDescriptor(True)
    """Generally you want to send these, lots of things will break if you don't send at least one."""
    AUTO_REQUEST_PARCELS: bool = SettingDescriptor(True)
    """Automatically request all parcel details when connecting to a region"""
    AUTO_REQUEST_MATERIALS: bool = SettingDescriptor(True)
    """Automatically request all materials when connecting to a region"""


class HippoCapsClient(CapsClient):
    def __init__(
            self,
            settings: ClientSettings,
            caps: Optional[CAPS_DICT] = None,
            session: Optional[aiohttp.ClientSession] = None,
    ) -> None:
        super().__init__(caps, session)
        self._settings = settings

    def _request_fixups(self, cap_or_url: str, headers: Dict, proxy: Optional[bool], ssl: Any):
        headers["User-Agent"] = self._settings.USER_AGENT
        return cap_or_url, headers, proxy, self._settings.SSL_VERIFY


class HippoClientProtocol(asyncio.DatagramProtocol):
    def __init__(self, session: HippoClientSession):
        self.session = proxify(session)
        self.message_xml = MessageDotXML()
        self.deserializer = UDPMessageDeserializer(
            settings=self.session.session_manager.settings,
        )

    def datagram_received(self, data, source_addr: ADDR_TUPLE):
        region = self.session.region_by_circuit_addr(source_addr)
        if not region:
            logging.warning("Received packet from invalid address %s", source_addr)
            return

        message = self.deserializer.deserialize(data)
        message.direction = Direction.IN
        message.sender = source_addr

        if not self.message_xml.validate_udp_msg(message.name):
            LOG.warning(
                f"Received {message.name!r} over UDP, when it should come over the event queue. Discarding."
            )
            raise PermissionError(f"UDPBanned message {message.name}")

        region.circuit.collect_acks(message)

        should_handle = True
        if message.reliable:
            # This is a bit crap. We send an ACK immediately through a PacketAck.
            # This is pretty wasteful, we should batch them up and send them on a timer.
            # We should ACK even if it's a resend of something we've already handled, maybe
            # they never got the ACK.
            region.circuit.send_acks((message.packet_id,))
            should_handle = region.circuit.track_reliable(message.packet_id)

        try:
            if should_handle:
                self.session.message_handler.handle(message)
        except:
            LOG.exception("Failed in session message handler")
        if should_handle:
            region.message_handler.handle(message)


class HippoClientRegion(BaseClientRegion):
    def __init__(self, circuit_addr, seed_cap: Optional[str], session: HippoClientSession, handle=None):
        super().__init__()
        self.caps = multidict.MultiDict()
        self.message_handler: MessageHandler[Message, str] = MessageHandler(take_by_default=False)
        self.circuit_addr = circuit_addr
        self.handle = handle
        if seed_cap:
            self.caps["Seed"] = seed_cap
        self.session: Callable[[], HippoClientSession] = weakref.ref(session)
        self.caps_client = HippoCapsClient(session.session_manager.settings, self.caps, session.http_session)
        self.xfer_manager = XferManager(proxify(self), self.session().secure_session_id)
        self.transfer_manager = TransferManager(proxify(self), session.agent_id, session.id)
        self.asset_uploader = AssetUploader(proxify(self))
        self.parcel_manager = ParcelManager(proxify(self))
        self.objects = ClientObjectManager(self)
        self._llsd_serializer = LLSDMessageSerializer()
        self._eq_task: Optional[asyncio.Task] = None
        self.connected: asyncio.Future = asyncio.Future()

        self.message_handler.subscribe("StartPingCheck", self._handle_ping_check)

    def update_caps(self, caps: Mapping[str, str]) -> None:
        self.caps.update(caps)

    @property
    def cap_urls(self) -> multidict.MultiDict:
        return self.caps.copy()

    async def connect(self, main_region: bool = False):
        # Disconnect first if we're already connected
        if self.circuit and self.circuit.is_alive:
            self.disconnect()
        if self.connected.done():
            self.connected = asyncio.Future()

        try:
            # TODO: What happens if a circuit code is invalid, again? Does it just refuse to ACK?
            await self.circuit.send_reliable(
                Message(
                    "UseCircuitCode",
                    Block(
                        "CircuitCode",
                        Code=self.session().circuit_code,
                        SessionID=self.session().id,
                        ID=self.session().agent_id,
                    ),
                )
            )
            self.circuit.is_alive = True

            # Clear out any old caps urls except the seed URL, we're about to fetch new caps.
            seed_url = self.caps["Seed"]
            self.caps.clear()
            self.caps["Seed"] = seed_url

            # Kick this off and await it later
            seed_resp_fut = self.caps_client.post("Seed", llsd=list(self.session().session_manager.SUPPORTED_CAPS))

            # Register first so we can handle it even if the ack happens after the message is sent
            region_handshake_fut = self.message_handler.wait_for(("RegionHandshake",))

            # If we're connecting to the main region, it won't even send us a RegionHandshake until we
            # first send a CompleteAgentMovement.
            if main_region:
                await self.complete_agent_movement()

            self.name = str((await region_handshake_fut)["RegionInfo"][0]["SimName"])
            self.session().objects.track_region_objects(self.handle)
            await self.circuit.send_reliable(
                Message(
                    "RegionHandshakeReply",
                    Block("AgentData", AgentID=self.session().agent_id, SessionID=self.session().id),
                    Block(
                        "RegionInfo",
                        Flags=(
                            RegionHandshakeReplyFlags.SUPPORTS_SELF_APPEARANCE
                            | RegionHandshakeReplyFlags.VOCACHE_CULLING_ENABLED
                        )
                    )
                )
            )
            await self.circuit.send_reliable(
                Message(
                    "AgentThrottle",
                    Block(
                        "AgentData",
                        AgentID=self.session().agent_id,
                        SessionID=self.session().id,
                        CircuitCode=self.session().circuit_code,
                    ),
                    Block(
                        "Throttle",
                        GenCounter=0,
                        # Reasonable defaults, I guess
                        Throttles_=ThrottleData(
                            resend=207360.0,
                            land=165376.0,
                            wind=33075.19921875,
                            cloud=33075.19921875,
                            task=682700.75,
                            texture=682700.75,
                            asset=269312.0
                        ),
                    )
                )
            )
            if self.session().session_manager.settings.SEND_AGENT_UPDATES:
                # Usually we want to send at least one, since lots of messages will never be sent by the sim
                # until we send at least one AgentUpdate. For example, ParcelOverlay and LayerData.
                await self.circuit.send_reliable(
                    Message(
                        "AgentUpdate",
                        Block(
                            'AgentData',
                            AgentID=self.session().agent_id,
                            SessionID=self.session().id,
                            # Don't really care about the other fields.
                            fill_missing=True,
                        )
                    )
                )

            async with seed_resp_fut as seed_resp:
                seed_resp.raise_for_status()
                self.update_caps(await seed_resp.read_llsd())

            self._eq_task = create_logged_task(self._poll_event_queue(), "EQ Poll")

            settings = self.session().session_manager.settings
            if settings.AUTO_REQUEST_PARCELS:
                _ = create_logged_task(self.parcel_manager.request_dirty_parcels(), "Parcel Request")
            if settings.AUTO_REQUEST_MATERIALS:
                _ = create_logged_task(self.objects.request_all_materials(), "Request All Materials")

        except Exception as e:
            # Let consumers who were `await`ing the connected signal know there was an error
            if not self.connected.done():
                self.connected.set_exception(e)
            raise

        self.connected.set_result(None)

    def disconnect(self) -> None:
        """Simulator has gone away, disconnect. Should be synchronous"""
        if self._eq_task is not None:
            self._eq_task.cancel()
        self._eq_task = None
        self.circuit.disconnect()
        self.objects.clear()
        if self.connected.done():
            self.connected = asyncio.Future()
        # TODO: cancel XFers and Transfers and whatnot

    async def complete_agent_movement(self) -> None:
        await self.circuit.send_reliable(
            Message(
                "CompleteAgentMovement",
                Block(
                    "AgentData",
                    AgentID=self.session().agent_id,
                    SessionID=self.session().id,
                    CircuitCode=self.session().circuit_code
                ),
            )
        )
        self.session().main_region = self

    async def _poll_event_queue(self):
        ack: Optional[int] = None
        while True:
            payload = {"ack": ack, "done": False}
            try:
                async with self.caps_client.post("EventQueueGet", llsd=payload) as resp:
                    if resp.status != 200:
                        await asyncio.sleep(0.1)
                        continue
                    polled = await resp.read_llsd()
                    for event in polled["events"]:
                        if self._llsd_serializer.can_handle(event["message"]):
                            msg = self._llsd_serializer.deserialize(event)
                        else:
                            msg = Message.from_eq_event(event)
                        msg.sender = self.circuit_addr
                        msg.direction = Direction.IN
                        self.session().message_handler.handle(msg)
                        self.message_handler.handle(msg)
                    ack = polled["id"]
                    await asyncio.sleep(0.001)
            except aiohttp.client_exceptions.ServerDisconnectedError:
                # This is expected to happen during long-polling, just pick up again where we left off.
                await asyncio.sleep(0.001)

    async def _handle_ping_check(self, message: Message):
        self.circuit.send(
            Message(
                "CompletePingCheck",
                Block("PingID", PingID=message["PingID"]["PingID"]),
            )
        )


class HippoClientSession(BaseClientSession):
    """Represents a client's view of a remote session"""
    REGION_CLS = HippoClientRegion

    region_by_handle: Callable[[int], Optional[HippoClientRegion]]
    region_by_circuit_addr: Callable[[ADDR_TUPLE], Optional[HippoClientRegion]]
    regions: List[HippoClientRegion]
    session_manager: HippoClient
    main_region: Optional[HippoClientRegion]

    def __init__(self, id, secure_session_id, agent_id, circuit_code, session_manager: Optional[HippoClient] = None,
                 login_data=None):
        super().__init__(id, secure_session_id, agent_id, circuit_code, session_manager, login_data=login_data)
        self.http_session = session_manager.http_session
        self.objects = ClientWorldObjectManager(proxify(self), session_manager.settings, None)
        self.inventory = InventoryManager(proxify(self))
        self.transport: Optional[SocketUDPTransport] = None
        self.protocol: Optional[HippoClientProtocol] = None
        self.message_handler.take_by_default = False

        for msg_name in ("DisableSimulator", "CloseCircuit"):
            self.message_handler.subscribe(msg_name, lambda msg: self.unregister_region(msg.sender))
        for msg_name in ("TeleportFinish", "CrossedRegion", "EstablishAgentCommunication"):
            self.message_handler.subscribe(msg_name, self._handle_register_region_message)

    def register_region(self, circuit_addr: Optional[ADDR_TUPLE] = None, seed_url: Optional[str] = None,
                        handle: Optional[int] = None) -> HippoClientRegion:
        return super().register_region(circuit_addr, seed_url, handle)  # type:ignore

    def unregister_region(self, circuit_addr: ADDR_TUPLE) -> None:
        for i, region in enumerate(self.regions):
            if region.circuit_addr == circuit_addr:
                self.regions[i].disconnect()
                del self.regions[i]
                return
        raise KeyError(f"No such region for {circuit_addr!r}")

    def open_circuit(self, circuit_addr: ADDR_TUPLE):
        for region in self.regions:
            if region.circuit_addr == circuit_addr:
                valid_circuit = False
                if not region.circuit or not region.circuit.is_alive:
                    region.circuit = Circuit(("127.0.0.1", 0), circuit_addr, self.transport)
                    region.circuit.is_alive = False
                    valid_circuit = True
                if region.circuit and region.circuit.is_alive:
                    # Whatever, already open
                    logging.debug("Tried to re-open circuit for %r" % (circuit_addr,))
                    valid_circuit = True
                return valid_circuit
        return False

    def _handle_register_region_message(self, msg: Message):
        # Handle events that inform us about new regions
        sim_addr, sim_handle, sim_seed = None, None, None
        moving_to_region = False
        # Sim is asking us to talk to a neighbour
        if msg.name == "EstablishAgentCommunication":
            ip_split = msg["EventData"]["sim-ip-and-port"].split(":")
            sim_addr = (ip_split[0], int(ip_split[1]))
            sim_seed = msg["EventData"]["seed-capability"]
        # We teleported or cross region, opening comms to new sim
        elif msg.name in ("TeleportFinish", "CrossedRegion"):
            sim_block = msg.get_blocks("RegionData", msg.get_blocks("Info"))[0]
            sim_addr = (sim_block["SimIP"], sim_block["SimPort"])
            sim_handle = sim_block["RegionHandle"]
            sim_seed = sim_block["SeedCapability"]
            moving_to_region = True
        # Sim telling us about a neighbour
        # elif msg.name == "EnableSimulator":
        #     sim_block = msg["SimulatorInfo"][0]
        #     sim_addr = (sim_block["IP"], sim_block["Port"])
        #     sim_handle = sim_block["Handle"]
        # TODO: EnableSimulator is a little weird. It creates a region and establishes a
        #  circuit, but with no seed cap. The viewer will send UseCircuitCode and all that,
        #  but it's totally workable to just wait for an EstablishAgentCommunication to do that,
        #  since that's when the region actually shows up. I guess EnableSimulator just gives the
        #  viewer some lead time to set up the circuit before the region is actually shown through
        #  EstablishAgentCommunication? Either way, messing around with regions that don't have seed
        #  caps is annoying, so let's just not do it.

        # Register a region if this message was telling us about a new one
        if sim_addr is not None:
            region = self.register_region(sim_addr, handle=sim_handle, seed_url=sim_seed)
            # We can't actually connect without a sim seed, mind you, when we receive and EnableSimulator
            # we have to wait for the EstablishAgentCommunication to actually connect.
            need_connect = (region.circuit and region.circuit.is_alive) or moving_to_region
            self.open_circuit(sim_addr)
            if need_connect:
                create_logged_task(region.connect(main_region=moving_to_region), "Region Connect")
            elif moving_to_region:
                # No need to connect, but we do need to complete agent movement.
                create_logged_task(region.complete_agent_movement(), "CompleteAgentMovement")


class HippoClient(BaseClientSessionManager):
    """A simple client, only connects to one region at a time currently."""

    SUPPORTED_CAPS: Set[str] = {
        "AbuseCategories",
        "AcceptFriendship",
        "AcceptGroupInvite",
        "AgentPreferences",
        "AgentProfile",
        "AgentState",
        "AttachmentResources",
        "AvatarPickerSearch",
        "AvatarRenderInfo",
        "CharacterProperties",
        "ChatSessionRequest",
        "CopyInventoryFromNotecard",
        "CreateInventoryCategory",
        "DeclineFriendship",
        "DeclineGroupInvite",
        "DispatchRegionInfo",
        "DirectDelivery",
        "EnvironmentSettings",
        "EstateAccess",
        "DispatchOpenRegionSettings",
        "EstateChangeInfo",
        "EventQueueGet",
        "ExtEnvironment",
        "FetchLib2",
        "FetchLibDescendents2",
        "FetchInventory2",
        "FetchInventoryDescendents2",
        "IncrementCOFVersion",
        "InventoryAPIv3",
        "LibraryAPIv3",
        "InterestList",
        "InventoryThumbnailUpload",
        "GetDisplayNames",
        "GetExperiences",
        "AgentExperiences",
        "FindExperienceByName",
        "GetExperienceInfo",
        "GetAdminExperiences",
        "GetCreatorExperiences",
        "ExperiencePreferences",
        "GroupExperiences",
        "UpdateExperience",
        "IsExperienceAdmin",
        "IsExperienceContributor",
        "RegionExperiences",
        "ExperienceQuery",
        "GetMesh",
        "GetMesh2",
        "GetMetadata",
        "GetObjectCost",
        "GetObjectPhysicsData",
        "GetTexture",
        "GroupAPIv1",
        "GroupMemberData",
        "GroupProposalBallot",
        "HomeLocation",
        "LandResources",
        "LSLSyntax",
        "MapLayer",
        "MapLayerGod",
        "MeshUploadFlag",
        "NavMeshGenerationStatus",
        "NewFileAgentInventory",
        "ObjectAnimation",
        "ObjectMedia",
        "ObjectMediaNavigate",
        "ObjectNavMeshProperties",
        "ParcelPropertiesUpdate",
        "ParcelVoiceInfoRequest",
        "ProductInfoRequest",
        "ProvisionVoiceAccountRequest",
        "ReadOfflineMsgs",
        "RegionObjects",
        "RemoteParcelRequest",
        "RenderMaterials",
        "RequestTextureDownload",
        "ResourceCostSelected",
        "RetrieveNavMeshSrc",
        "SearchStatRequest",
        "SearchStatTracking",
        "SendPostcard",
        "SendUserReport",
        "SendUserReportWithScreenshot",
        "ServerReleaseNotes",
        "SetDisplayName",
        "SimConsoleAsync",
        "SimulatorFeatures",
        "StartGroupProposal",
        "TerrainNavMeshProperties",
        "TextureStats",
        "UntrustedSimulatorMessage",
        "UpdateAgentInformation",
        "UpdateAgentLanguage",
        "UpdateAvatarAppearance",
        "UpdateGestureAgentInventory",
        "UpdateGestureTaskInventory",
        "UpdateNotecardAgentInventory",
        "UpdateNotecardTaskInventory",
        "UpdateScriptAgent",
        "UpdateScriptTask",
        "UpdateSettingsAgentInventory",
        "UpdateSettingsTaskInventory",
        "UploadAgentProfileImage",
        "UploadBakedTexture",
        "UserInfo",
        "ViewerAsset",
        "ViewerBenefits",
        "ViewerMetrics",
        "ViewerStartAuction",
        "ViewerStats",
    }

    DEFAULT_OPTIONS = {
        "inventory-root",
        "inventory-skeleton",
        "inventory-lib-root",
        "inventory-lib-owner",
        "inventory-skel-lib",
        "initial-outfit",
        "gestures",
        "display_names",
        "event_notifications",
        "classified_categories",
        "adult_compliant",
        "buddy-list",
        "newuser-config",
        "ui-config",
        "advanced-mode",
        "max-agent-groups",
        "map-server-url",
        "voice-config",
        "tutorial_setting",
        "login-flags",
        "global-textures",
        # Not an official option, just so this can be tracked.
        "pyogp-client",
    }

    DEFAULT_LOGIN_URI = "https://login.agni.lindenlab.com/cgi-bin/login.cgi"

    def __init__(self, options: Optional[Set[str]] = None):
        self._username: Optional[str] = None
        self._password: Optional[str] = None
        self._mac = uuid.getnode()
        self._options = options if options is not None else self.DEFAULT_OPTIONS
        self.http_session: Optional[aiohttp.ClientSession] = aiohttp.ClientSession(trust_env=True)
        self.session: Optional[HippoClientSession] = None
        self.settings = ClientSettings()
        self._resend_task: Optional[asyncio.Task] = None

    @property
    def main_region(self) -> Optional[HippoClientRegion]:
        if not self.session:
            return None
        return self.session.main_region

    @property
    def main_circuit(self) -> Optional[Circuit]:
        if not self.main_region:
            return None
        return self.main_region.circuit

    @property
    def main_caps_client(self) -> Optional[CapsClient]:
        if not self.main_region:
            return None
        return self.main_region.caps_client

    async def aclose(self):
        try:
            self.logout()
        finally:
            if self.http_session:
                await self.http_session.close()
                self.http_session = None

    def __del__(self):
        # Make sure we don't leak resources if someone was lazy.
        try:
            self.logout()
        finally:
            if self.http_session:
                try:
                    asyncio.create_task(self.http_session.close)
                except:
                    pass
                self.http_session = None

    async def _create_transport(self) -> Tuple[AbstractUDPTransport, HippoClientProtocol]:
        loop = asyncio.get_event_loop_policy().get_event_loop()
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: HippoClientProtocol(self.session),
            local_addr=('0.0.0.0', 0))
        transport = SocketUDPTransport(transport)
        return transport, protocol

    async def login(
            self,
            username: str,
            password: str,
            login_uri: Optional[str] = None,
            agree_to_tos: bool = False,
            start_location: Union[StartLocation, str, None] = StartLocation.LAST,
            connect: bool = True,
    ):
        if self.session:
            raise RuntimeError("Already logged in!")

        if not login_uri:
            login_uri = self.DEFAULT_LOGIN_URI

        if start_location is None:
            start_location = StartLocation.LAST

        # This isn't a symbolic start location and isn't a URI, must be a sim name.
        if start_location not in iter(StartLocation) and not start_location.startswith("uri:"):
            start_location = f"uri:{start_location}&128&128&128"

        split_username = username.split(" ")
        if len(split_username) < 2:
            first_name = split_username[0]
            last_name = "Resident"
        else:
            first_name, last_name = split_username

        payload = {
            "address_size": 64,
            "agree_to_tos": int(agree_to_tos),
            "channel": "Hippolyzer",
            "extended_errors": 1,
            "first": first_name,
            "last": last_name,
            "host_id": "",
            "id0": hashlib.md5(str(self._mac).encode("ascii")).hexdigest(),
            "mac": hashlib.md5(str(self._mac).encode("ascii")).hexdigest(),
            "mfa_hash": "",
            "passwd": "$1$" + hashlib.md5(str(password).encode("ascii")).hexdigest(),
            # TODO: actually get these
            "platform": "lnx",
            "platform_string": "Linux 6.6",
            # TODO: What is this?
            "platform_version": "2.38.0",
            "read_critical": 0,
            "start": str(start_location),
            "token": "",
            "version": version("hippolyzer"),
            "options": list(self._options),
        }
        async with self.http_session.post(
                login_uri,
                data=xmlrpc.client.dumps((payload,), "login_to_simulator"),
                headers={"Content-Type": "text/xml", "User-Agent": self.settings.USER_AGENT},
                ssl=self.settings.SSL_VERIFY,
        ) as resp:
            resp.raise_for_status()
            login_data = xmlrpc.client.loads((await resp.read()).decode("utf8"))[0][0]
        self.session = HippoClientSession.from_login_data(login_data, self)

        self.session.transport, self.session.protocol = await self._create_transport()
        self._resend_task = create_logged_task(self._attempt_resends(), "Circuit Resend")
        self.session.message_handler.subscribe("AgentDataUpdate", self._handle_agent_data_update)
        self.session.message_handler.subscribe("AgentGroupDataUpdate", self._handle_agent_group_data_update)

        assert self.session.open_circuit(self.session.regions[-1].circuit_addr)
        if connect:
            region = self.session.regions[-1]
            await region.connect(main_region=True)

    def logout(self):
        if not self.session:
            return
        if self._resend_task:
            self._resend_task.cancel()
            self._resend_task = None

        if self.main_circuit and self.main_circuit.is_alive:
            # Don't need to send reliably, there's a good chance the server won't ACK anyway.
            self.main_circuit.send(
                Message(
                    "LogoutRequest",
                    Block("AgentData", AgentID=self.session.agent_id, SessionID=self.session.id),
                )
            )
        session = self.session
        self.session = None
        for region in session.regions:
            region.disconnect()
        session.transport.close()

    def send_chat(self, message: Union[bytes, str], channel: int = 0, chat_type=ChatType.NORMAL) -> asyncio.Future:
        return self.main_circuit.send_reliable(Message(
            "ChatFromViewer",
            Block("AgentData", SessionID=self.session.id, AgentID=self.session.agent_id),
            Block("ChatData", Message=message, Channel=channel, Type=chat_type),
        ))

    def teleport(self, region_handle: int, local_pos=Vector3(0, 0, 0)) -> asyncio.Future:
        """Synchronously requests a teleport, returning a Future for teleport completion"""
        teleport_fut = asyncio.Future()

        # Send request synchronously, await asynchronously.
        send_fut = self.main_circuit.send_reliable(
            Message(
                'TeleportLocationRequest',
                Block('AgentData', AgentID=self.session.agent_id, SessionID=self.session.id),
                Block('Info', RegionHandle=region_handle, Position=local_pos, fill_missing=True),
            )
        )

        async def _handle_teleport():
            # Subscribe first, we may receive an event before we receive the packet ACK.
            with self.session.message_handler.subscribe_async(
                    ("TeleportLocal", "TeleportFailed", "TeleportFinish"),
            ) as get_tp_done_msg:
                try:
                    await send_fut
                except Exception as e:
                    # Pass along error if we failed to send reliably.
                    teleport_fut.set_exception(e)
                    return

                # Wait for a message that says we're done the teleport
                msg = await get_tp_done_msg()
                if msg.name == "TeleportFailed":
                    teleport_fut.set_exception(RuntimeError("Failed to teleport"))
                elif msg.name == "TeleportLocal":
                    # Within the sim, nothing else we need to do
                    teleport_fut.set_result(None)
                elif msg.name == "TeleportFinish":
                    # Non-local TP, wait until we receive the AgentMovementComplete to
                    # set the finished signal.

                    # Region should be registered by this point, wait for it to connect
                    try:
                        # just fail if it takes longer than 30 seconds for the handshake to complete
                        await asyncio.wait_for(self.session.region_by_handle(region_handle).connected, 30)
                    except Exception as e:
                        teleport_fut.set_exception(e)
                        return
                    teleport_fut.set_result(None)

        create_logged_task(_handle_teleport(), "Teleport")

        return teleport_fut

    async def _attempt_resends(self):
        while True:
            if self.session is None:
                break
            for region in self.session.regions:
                if not region.circuit.is_alive:
                    continue
                region.circuit.resend_unacked()
            await asyncio.sleep(0.5)

    def _handle_agent_data_update(self, msg: Message):
        self.session.active_group = msg["AgentData"]["ActiveGroupID"]

    def _handle_agent_group_data_update(self, msg: Message):
        self.session.groups.clear()
        for block in msg["GroupData"]:
            self.session.groups.add(block["GroupID"])



"""
Example of how to request an Xfer
"""
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.inventory import InventoryModel
from hippolyzer.lib.base.templates import XferFilePath, AssetType, InventoryType, WearableType
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.proxy.addon_utils import BaseAddon, show_message
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class XferExampleAddon(BaseAddon):
    @handle_command()
    async def get_mute_list(self, session: Session, region: ProxiedRegion):
        """Fetch the current user's mute list"""
        region.circuit.send(Message(
            'MuteListRequest',
            Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
            Block("MuteData", MuteCRC=0),
        ))

        # Wait for any MuteListUpdate, dropping it before it reaches the viewer
        update_msg = await region.message_handler.wait_for(('MuteListUpdate',), timeout=5.0)
        mute_file_name = update_msg["MuteData"]["Filename"]
        if not mute_file_name:
            show_message("Nobody muted?")
            return

        xfer = await region.xfer_manager.request(
            file_name=mute_file_name, file_path=XferFilePath.CACHE)
        show_message(xfer.reassemble_chunks().decode("utf8"))

    @handle_command()
    async def get_task_inventory(self, session: Session, region: ProxiedRegion):
        """Get the inventory of the currently selected object"""
        region.circuit.send(Message(
            'RequestTaskInventory',
            # If no session is passed in we'll use the active session when the coro was created
            Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
            Block('InventoryData', LocalID=session.selected.object_local),
        ))

        inv_message = await region.message_handler.wait_for(('ReplyTaskInventory',), timeout=5.0)

        # Xfer doesn't need to be immediately awaited, multiple signals can be waited on.
        xfer = region.xfer_manager.request(
            file_name=inv_message["InventoryData"]["Filename"], file_path=XferFilePath.CACHE)

        # Wait until we have the first packet so we can tell the expected length
        # The difference in time is obvious for large inventories, and we can cancel
        # mid-request if we want.
        show_message(f"Inventory is {await xfer.size_known} bytes")

        # Wait for the rest of the body to be done
        await xfer

        inv_model = InventoryModel.from_bytes(xfer.reassemble_chunks())
        item_names = [item.name for item in inv_model.all_items]
        show_message(item_names)

    @handle_command()
    async def eyes_for_you(self, session: Session, region: ProxiedRegion):
        """Upload an eye bodypart and create an item for it"""
        asset_data = f"""LLWearable version 22
New Eyes

\tpermissions 0
\t{{
\t\tbase_mask\t7fffffff
\t\towner_mask\t7fffffff
\t\tgroup_mask\t00000000
\t\teveryone_mask\t00000000
\t\tnext_owner_mask\t00082000
\t\tcreator_id\t{session.agent_id}
\t\towner_id\t{session.agent_id}
\t\tlast_owner_id\t00000000-0000-0000-0000-000000000000
\t\tgroup_id\t00000000-0000-0000-0000-000000000000
\t}}
\tsale_info\t0
\t{{
\t\tsale_type\tnot
\t\tsale_price\t10
\t}}
type 3
parameters 2
98 0
99 0
textures 1
3 89556747-24cb-43ed-920b-47caed15465f
"""
        # If we want to create an item containing the asset we need to know the transaction id
        # used to create the asset.
        transaction_id = UUID.random()
        await region.xfer_manager.upload_asset(
            AssetType.BODYPART,
            data=asset_data,
            transaction_id=transaction_id
        )
        region.circuit.send(Message(
            'CreateInventoryItem',
            Block('AgentData', AgentID=session.agent_id, SessionID=session.id),
            Block(
                'InventoryBlock',
                CallbackID=0,
                # Null folder ID will put it in the default folder for the type
                FolderID=UUID(),
                TransactionID=transaction_id,
                NextOwnerMask=0x7fFFffFF,
                Type=AssetType.BODYPART,
                InvType=InventoryType.WEARABLE,
                WearableType=WearableType.EYES,
                Name='Eyes For You',
                Description=b''
            ),
        ))


addons = [XferExampleAddon()]



import unittest

from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.msgtypes import PacketFlags
from hippolyzer.lib.proxy.circuit import ProxiedCircuit, InjectionTracker
from hippolyzer.lib.base.network.transport import Direction


class MockedProxyCircuit(ProxiedCircuit):
    def __init__(self):
        super().__init__(("127.0.0.1", 0), ("0.0.0.0", 0), None)
        self.sent_simple = []
        self.sent_msgs = []
        # Use a small maxlen to test what happens when injected
        # IDs get evicted from the deque
        self.out_injections = InjectionTracker(0, maxlen=10)
        self.in_injections = InjectionTracker(0, maxlen=10)

    def _send_prepared_message(self, msg: Message, transport=None):
        self.sent_simple.append((msg.packet_id, msg.name, msg.direction, msg.synthetic, msg.acks))
        self.sent_msgs.append(msg)


class PacketIDTests(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.circuit = MockedProxyCircuit()

    def _send_message(self, msg, outgoing=True):
        msg.direction = Direction.OUT if outgoing else Direction.IN
        return self.circuit.send(msg)

    async def test_basic(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer', packet_id=2))

        self.assertSequenceEqual(self.circuit.sent_simple, (
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, False, ()),
        ))

    async def test_inject(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', packet_id=2))

        self.assertSequenceEqual(self.circuit.sent_simple, (
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, True, ()),
            (3, "ChatFromViewer", Direction.OUT, False, ()),
        ))

    async def test_max_injected(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        for _ in range(5):
            self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', packet_id=2))

        self.assertEqual(self.circuit.out_injections.get_original_id(1), 1)
        self.assertEqual(self.circuit.out_injections.get_original_id(7), 2)

        for _ in range(7):
            self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', packet_id=3))

        self.assertEqual(len(self.circuit.sent_simple), 15)

        for i in range(15):
            pack = self.circuit.sent_simple[i]
            # Check that the sequence nums match what we'd expect
            self.assertEqual(pack[0], i + 1)
            # Make sure only 1 and 15 are marked not injected
            self.assertEqual(pack[3], i + 1 not in (1, 7, 15))

        # Make sure we're still able to get the original ID
        self.assertEqual(self.circuit.out_injections.get_original_id(15), 3)

    async def test_inject_hole_in_sequence(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', packet_id=4))
        self._send_message(Message('ChatFromViewer'))

        self.assertSequenceEqual(self.circuit.sent_simple, (
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, True, ()),
            (5, "ChatFromViewer", Direction.OUT, False, ()),
            (6, "ChatFromViewer", Direction.OUT, True, ()),
        ))

    async def test_inject_misordered(self):
        self._send_message(Message('ChatFromViewer', packet_id=2))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', packet_id=1))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (2, "ChatFromViewer", Direction.OUT, False, ()),
            (3, "ChatFromViewer", Direction.OUT, True, ()),
            (1, "ChatFromViewer", Direction.OUT, False, ()),
        ])

    async def test_inject_multiple(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', packet_id=2))
        self._send_message(Message('ChatFromViewer'))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, True, ()),
            (3, "ChatFromViewer", Direction.OUT, True, ()),
            (4, "ChatFromViewer", Direction.OUT, True, ()),
            (5, "ChatFromViewer", Direction.OUT, False, ()),
            (6, "ChatFromViewer", Direction.OUT, True, ()),
        ])

    async def test_packet_ack_field_converted(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', flags=PacketFlags.RELIABLE))
        self._send_message(Message('ChatFromSimulator', packet_id=1, acks=(2, 3, 4)), outgoing=False)
        self._send_message(Message('ChatFromViewer', packet_id=2, flags=PacketFlags.RELIABLE))
        self._send_message(Message('ChatFromSimulator', packet_id=2, acks=[5]), outgoing=False)
        self._send_message(Message('ChatFromViewer'))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, True, ()),
            (3, "ChatFromViewer", Direction.OUT, True, ()),
            (4, "ChatFromViewer", Direction.OUT, True, ()),
            # Ack for packet viewer didn't know about should be removed
            (1, "ChatFromSimulator", Direction.IN, False, ()),
            (5, "ChatFromViewer", Direction.OUT, False, ()),
            # Ack for packet viewer did know about should have its ID shifted down
            # to account for injections viewer didn't know about
            (2, "ChatFromSimulator", Direction.IN, False, (2,)),
            (6, "ChatFromViewer", Direction.OUT, True, ()),
        ])

    async def test_packet_ack_proxied_message_converted(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer'))
        self._send_message(Message('ChatFromViewer', flags=PacketFlags.RELIABLE))
        self._send_message(
            Message(
                'PacketAck',
                Block('Packets', ID=2),
                Block('Packets', ID=3),
                Block('Packets', ID=4),
                packet_id=1,
            ),
            outgoing=False
        )
        self._send_message(Message('ChatFromViewer', packet_id=2, flags=PacketFlags.RELIABLE))
        self._send_message(
            Message('PacketAck', Block('Packets', ID=5), packet_id=2),
            outgoing=False
        )
        self._send_message(Message('ChatFromViewer'))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, True, ()),
            (3, "ChatFromViewer", Direction.OUT, True, ()),
            (4, "ChatFromViewer", Direction.OUT, True, ()),
            # Ack for packet viewer didn't know about was dropped, not present
            (5, "ChatFromViewer", Direction.OUT, False, ()),
            # Ack for packet viewer did know about should have its ID shifted down
            # to account for injections viewer didn't know about
            (2, "PacketAck", Direction.IN, False, ()),
            (6, "ChatFromViewer", Direction.OUT, True, ()),
        ])

        self.assertEqual(self.circuit.sent_msgs[5]["Packets"][0]["ID"], 2)

    async def test_drop_proxied_message(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self.circuit.drop_message(Message('ChatFromViewer', packet_id=2, flags=PacketFlags.RELIABLE))
        self._send_message(Message('ChatFromViewer', packet_id=3))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (1, "PacketAck", Direction.IN, True, ()),
            (3, "ChatFromViewer", Direction.OUT, False, ()),
        ])
        self.assertEqual(self.circuit.sent_msgs[1]["Packets"][0]["ID"], 2)

    async def test_unreliable_proxied_message(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self.circuit.drop_message(Message('ChatFromViewer', packet_id=2))
        self._send_message(Message('ChatFromViewer', packet_id=3))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (3, "ChatFromViewer", Direction.OUT, False, ()),
        ])

    async def test_dropped_proxied_message_acks_sent(self):
        self._send_message(Message('ChatFromViewer', packet_id=1))
        self._send_message(Message('ChatFromViewer', packet_id=2))
        self._send_message(Message('ChatFromViewer', packet_id=3))
        self._send_message(Message('ChatFromSimulator'), outgoing=False)
        self.circuit.drop_message(Message('ChatFromViewer', packet_id=4, acks=(4,)))
        self._send_message(Message('ChatFromViewer', packet_id=5))

        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (2, "ChatFromViewer", Direction.OUT, False, ()),
            (3, "ChatFromViewer", Direction.OUT, False, ()),
            (1, "ChatFromSimulator", Direction.IN, True, ()),
            # PacketAcking embedded acks on message dropping is a
            # special case that will just use the dropped message's ID.
            # If it had no acks we just leave a hole in the message sequence.
            (4, "PacketAck", Direction.OUT, True, ()),
            (5, "ChatFromViewer", Direction.OUT, False, ()),
        ])
        # We injected an incoming packet, so "4" is really "3"
        self.assertEqual(self.circuit.sent_msgs[4]["Packets"][0]["ID"], 3)

    async def test_resending_or_dropping(self):
        self.circuit.send(Message('ChatFromViewer', packet_id=1))
        to_drop = Message('ChatFromViewer', packet_id=2, flags=PacketFlags.RELIABLE)
        self.circuit.drop_message(to_drop)
        with self.assertRaises(RuntimeError):
            # Re-dropping the same message should raise
            self.circuit.drop_message(to_drop)
        # Returns a new message without finalized flag
        new_msg = to_drop.take()
        self.circuit.send(new_msg)
        with self.assertRaises(RuntimeError):
            self.circuit.send(new_msg)
        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, False, ()),
            (1, "PacketAck", Direction.IN, True, ()),
            # ended up getting the same packet ID when injected
            (2, "ChatFromViewer", Direction.OUT, True, ()),
        ])

    async def test_reliable_unacked_queueing(self):
        self._send_message(Message('ChatFromViewer', flags=PacketFlags.RELIABLE))
        self._send_message(Message('ChatFromViewer', flags=PacketFlags.RELIABLE, packet_id=2))
        # Only the first, injected message should be queued for resends
        self.assertEqual({(Direction.OUT, 1)}, set(self.circuit.unacked_reliable))

    async def test_reliable_resend_cadence(self):
        self._send_message(Message('ChatFromViewer', flags=PacketFlags.RELIABLE))
        resend_info = self.circuit.unacked_reliable[(Direction.OUT, 1)]
        self.circuit.resend_unacked()
        # Should have been too soon to retry
        self.assertEqual(10, resend_info.tries_left)
        # Switch to allowing resends every 0s
        self.circuit.resend_every = 0.0
        self.circuit.resend_unacked()
        self.assertSequenceEqual(self.circuit.sent_simple, [
            (1, "ChatFromViewer", Direction.OUT, True, ()),
            # Should have resent
            (1, "ChatFromViewer", Direction.OUT, True, ()),
        ])
        self.assertEqual(9, resend_info.tries_left)
        for _ in range(resend_info.tries_left):
            self.circuit.resend_unacked()
        # Should have used up all the retry attempts and been kicked out of the retry queue
        self.assertEqual(set(), set(self.circuit.unacked_reliable))

    async def test_reliable_ack_collection(self):
        msg = Message('ChatFromViewer', flags=PacketFlags.RELIABLE)
        fut = self.circuit.send_reliable(msg)
        self.assertEqual(1, len(self.circuit.unacked_reliable))
        # Shouldn't count, this is an ACK going in the wrong direction!
        ack_msg = Message("PacketAck", Block("Packets", ID=msg.packet_id))
        self.circuit.collect_acks(ack_msg)
        self.assertEqual(1, len(self.circuit.unacked_reliable))
        self.assertFalse(fut.done())
        # But it should count if the ACK message is heading in
        ack_msg.direction = Direction.IN
        self.circuit.collect_acks(ack_msg)
        self.assertEqual(0, len(self.circuit.unacked_reliable))
        self.assertTrue(fut.done())

    async def test_start_ping_check(self):
        # Should not break if no unacked
        self._send_message(Message(
            "StartPingCheck",
            Block("PingID", PingID=0, OldestUnacked=20),
            packet_id=5,
        ))

        injected_msg = Message('ChatFromViewer', flags=PacketFlags.RELIABLE)
        self._send_message(injected_msg)

        self._send_message(Message(
            "StartPingCheck",
            Block("PingID", PingID=0, OldestUnacked=20),
            packet_id=8,
        ))
        # Oldest unacked should have been replaced with the injected packet's ID, it's older!
        self.assertEqual(self.circuit.sent_msgs[2]["PingID"]["OldestUnacked"], injected_msg.packet_id)



"""
Tail animation generator

Demonstrates programmatic generation of local motions using BaseAnimHelperAddon

You can use this to create an animation with a script, fiddle with it until it
looks right, then finally save it with /524 save_local_anim <ANIM_NAME>.

The built animation is automatically applied to all active sessions when loaded,
and is re-generated whenever the script is edited. Unloading the script stops
the animations.
"""

from hippolyzer.lib.base.anim_utils import shift_keyframes, smooth_rot
from hippolyzer.lib.base.datatypes import Quaternion
from hippolyzer.lib.base.llanim import Animation, Joint
from hippolyzer.lib.proxy.addons import AddonManager

import local_anim
AddonManager.hot_reload(local_anim, require_addons_loaded=True)


class TailAnimator(local_anim.BaseAnimHelperAddon):
    # Should be unique
    ANIM_NAME = "tail_anim"

    def build_anim(self) -> Animation:
        anim = Animation(
            base_priority=5,
            duration=5.0,
            loop_out_point=5.0,
            loop=True,
        )
        # Iterate along tail joints 1 through 6
        for joint_num in range(1, 7):
            # Give further along joints a wider range of motion
            start_rot = Quaternion.from_euler(0.2, -0.3, 0.15 * joint_num)
            end_rot = Quaternion.from_euler(-0.2, -0.3, -0.15 * joint_num)
            rot_keyframes = [
                # Tween between start_rot and end_rot, using smooth interpolation.
                # SL's keyframes only allow linear interpolation which doesn't look great
                # for natural motions. `smooth_rot()` gets around that by generating
                # smooth inter frames for SL to linearly interpolate between.
                *smooth_rot(start_rot, end_rot, inter_frames=10, time=0.0, duration=2.5),
                *smooth_rot(end_rot, start_rot, inter_frames=10, time=2.5, duration=2.5),
            ]
            anim.joints[f"mTail{joint_num}"] = Joint(
                priority=5,
                # Each joint's frames should be ahead of the previous joint's by 2 frames
                rot_keyframes=shift_keyframes(rot_keyframes, joint_num * 2),
            )
        return anim


addons = [TailAnimator()]



from __future__ import annotations

import copy
import dataclasses
import re
import weakref
from typing import *

import transformations
from lxml import etree

from hippolyzer.lib.base.datatypes import Vector3, RAD_TO_DEG
from hippolyzer.lib.base.helpers import get_resource_filename
from hippolyzer.lib.base.mesh import MeshAsset, SkinSegmentDict, llsd_to_mat4

MAYBE_JOINT_REF = Optional[str]
SKELETON_REF = Optional[Callable[[], "Skeleton"]]


@dataclasses.dataclass
class JointNode:
    name: str
    parent_name: MAYBE_JOINT_REF
    skeleton: SKELETON_REF
    translation: Vector3
    pivot: Vector3  # pivot point for the joint, generally the same as translation
    rotation: Vector3  # Euler rotation in degrees
    scale: Vector3
    type: str  # bone or collision_volume
    support: str

    def __hash__(self):
        return hash((self.name, self.type))

    @property
    def matrix(self):
        return transformations.compose_matrix(
            scale=tuple(self.scale),
            angles=tuple(self.rotation / RAD_TO_DEG),
            translate=tuple(self.translation),
        )

    @property
    def parent(self) -> Optional[JointNode]:
        if self.parent_name:
            return self.skeleton()[self.parent_name]
        return None

    @property
    def index(self) -> int:
        bone_idx = 0
        for node in self.skeleton().joint_dict.values():
            if node.type != "bone":
                continue
            if self is node:
                return bone_idx
            bone_idx += 1
        raise KeyError(f"{self.name!r} doesn't exist in skeleton")

    @property
    def ancestors(self) -> Sequence[JointNode]:
        joint_node = self
        skeleton = self.skeleton()
        ancestors: List[JointNode] = []
        while joint_node.parent_name:
            joint_node = skeleton.joint_dict.get(joint_node.parent_name)
            ancestors.append(joint_node)
        return ancestors

    @property
    def children(self) -> Sequence[JointNode]:
        children: List[JointNode] = []
        for node in self.skeleton().joint_dict.values():
            if node.parent_name and node.parent_name == self.name:
                children.append(node)
        return children

    @property
    def inverse(self) -> Optional[JointNode]:
        l_re = re.compile(r"(.*?(?:_|\b))L((?:_|\b).*)")
        r_re = re.compile(r"(.*?(?:_|\b))R((?:_|\b).*)")

        inverse_name = None
        if "Left" in self.name:
            inverse_name = self.name.replace("Left", "Right")
        elif "LEFT" in self.name:
            inverse_name = self.name.replace("LEFT", "RIGHT")
        elif l_re.match(self.name):
            inverse_name = re.sub(l_re, r"\1R\2", self.name)
        elif "Right" in self.name:
            inverse_name = self.name.replace("Right", "Left")
        elif "RIGHT" in self.name:
            inverse_name = self.name.replace("RIGHT", "LEFT")
        elif r_re.match(self.name):
            inverse_name = re.sub(r_re, r"\1L\2", self.name)

        if inverse_name:
            return self.skeleton().joint_dict.get(inverse_name)
        return None

    @property
    def descendents(self) -> Set[JointNode]:
        descendents: Set[JointNode] = set()
        ancestors: Set[str] = {self.name}
        last_ancestors: Set[str] = set()
        while last_ancestors != ancestors:
            last_ancestors = ancestors.copy()
            for node in self.skeleton().joint_dict.values():
                if node.parent_name and node.parent_name in ancestors:
                    ancestors.add(node.name)
                    descendents.add(node)
        return descendents


class Skeleton:
    def __init__(self, root_node: Optional[etree.ElementBase] = None):
        self.joint_dict: Dict[str, JointNode] = {}
        if root_node is not None:
            self._parse_node_children(root_node, None)

    def __getitem__(self, item: str) -> JointNode:
        return self.joint_dict[item]

    def clone(self) -> Self:
        val = copy.deepcopy(self)
        skel_ref = weakref.ref(val)
        for joint in val.joint_dict.values():
            joint.skeleton = skel_ref
        return val

    def _parse_node_children(self, node: etree.ElementBase, parent_name: MAYBE_JOINT_REF):
        name = node.get('name')
        joint = JointNode(
            name=name,
            parent_name=parent_name,
            skeleton=weakref.ref(self),
            translation=_get_vec_attr(node, "pos", Vector3()),
            pivot=_get_vec_attr(node, "pivot", Vector3()),
            rotation=_get_vec_attr(node, "rot", Vector3()),
            scale=_get_vec_attr(node, "scale", Vector3(1, 1, 1)),
            support=node.get('support', 'base'),
            type=node.tag,
        )
        self.joint_dict[name] = joint
        for child in node.iterchildren():
            self._parse_node_children(child, joint.name)

    def merge_mesh_skeleton(self, mesh: MeshAsset) -> None:
        """Update this skeleton with a skeleton definition from a mesh asset"""
        skin_seg: Optional[SkinSegmentDict] = mesh.segments.get('skin')
        if not skin_seg:
            return

        for joint_name, matrix in zip(skin_seg['joint_names'], skin_seg.get('alt_inverse_bind_matrix', [])):
            # We're only meant to use the translation component from the alt inverse bind matrix.
            joint_decomp = transformations.decompose_matrix(llsd_to_mat4(matrix))
            joint_node = self.joint_dict.get(joint_name)
            if not joint_node:
                continue
            joint_node.translation = Vector3(*joint_decomp[3])

        if pelvis_offset := skin_seg.get('pelvis_offset'):
            # TODO: Should we even do this?
            pelvis_node = self["mPelvis"]
            pelvis_node.translation += Vector3(0, 0, pelvis_offset)


def _get_vec_attr(node, attr_name: str, default: Vector3) -> Vector3:
    attr_val = node.get(attr_name, None)
    if not attr_val:
        return default
    return Vector3(*(float(x) for x in attr_val.split(" ") if x))


def load_avatar_skeleton() -> Skeleton:
    skel_path = get_resource_filename("lib/base/data/avatar_skeleton.xml")
    with open(skel_path, 'r') as f:
        skel_root = etree.fromstring(f.read())
    return Skeleton(skel_root.getchildren()[0])


AVATAR_SKELETON = load_avatar_skeleton()



"""
Check object manager state against region ViewerObject cache

Can't look at every object we've tracked and every object in VOCache
and report mismatches due to weird VOCache cache eviction criteria and certain
cacheable objects not being added to the VOCache.

Off the top of my head, animesh objects get explicit KillObjects at extreme
view distances same as avatars, but will still be present in the cache even
though they will not be in gObjectList.
"""
import asyncio
import logging
from typing import *

from hippolyzer.lib.base.objects import normalize_object_update_compressed_data
from hippolyzer.lib.base.templates import ObjectUpdateFlags, PCode
from hippolyzer.lib.proxy.addon_utils import BaseAddon, GlobalProperty
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import SessionManager, Session
from hippolyzer.lib.proxy.vocache import is_valid_vocache_dir, RegionViewerObjectCacheChain

LOG = logging.getLogger(__name__)


class ObjectManagementValidator(BaseAddon):
    base_cache_path: Optional[str] = GlobalProperty(None)
    orig_auto_request: Optional[bool] = GlobalProperty(None)

    def handle_init(self, session_manager: SessionManager):
        if self.orig_auto_request is None:
            self.orig_auto_request = session_manager.settings.ALLOW_AUTO_REQUEST_OBJECTS
        session_manager.settings.ALLOW_AUTO_REQUEST_OBJECTS = False

        async def _choose_cache_path():
            while not self.base_cache_path:
                cache_dir = await AddonManager.UI.open_dir("Choose the base cache directory")
                if not cache_dir:
                    return
                if not is_valid_vocache_dir(cache_dir):
                    continue
                self.base_cache_path = cache_dir

        if not self.base_cache_path:
            self._schedule_task(_choose_cache_path(), session_scoped=False)

    def handle_unload(self, session_manager: SessionManager):
        session_manager.settings.ALLOW_AUTO_REQUEST_OBJECTS = self.orig_auto_request

    def handle_session_init(self, session: Session):
        # Use only the specified cache path for the vocache
        session.cache_dir = self.base_cache_path

    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if message.name != "DisableSimulator":
            return
        # Send it off to the client without handling it normally,
        # we need to defer region teardown in the proxy
        region.circuit.send(message)
        self._schedule_task(self._check_cache_before_region_teardown(region))
        return True

    async def _check_cache_before_region_teardown(self, region: ProxiedRegion):
        await asyncio.sleep(0.5)
        print("Ok, checking cache differences")
        try:
            # Index will have been rewritten, so re-read it.
            region_cache_chain = RegionViewerObjectCacheChain.for_region(
                handle=region.handle,
                cache_id=region.cache_id,
                cache_dir=self.base_cache_path
            )
            if not region_cache_chain.region_caches:
                print(f"no caches for {region!r}?")
                return
            all_full_ids = set()
            for obj in region.objects.all_objects:
                cacheable = True
                orig_obj = obj
                # Walk along the ancestry checking for things that would make the tree non-cacheable
                while obj is not None:
                    if obj.UpdateFlags & ObjectUpdateFlags.TEMPORARY_ON_REZ:
                        cacheable = False
                    if obj.PCode == PCode.AVATAR:
                        cacheable = False
                    obj = obj.Parent
                if cacheable:
                    all_full_ids.add(orig_obj.FullID)

            for key in all_full_ids:
                obj = region.objects.lookup_fullid(key)
                cached_data = region_cache_chain.lookup_object_data(obj.LocalID, obj.CRC)
                if not cached_data:
                    continue
                orig_dict = obj.to_dict()
                parsed_data = normalize_object_update_compressed_data(cached_data)
                updated = obj.update_properties(parsed_data)
                # Can't compare this yet
                updated -= {"TextureEntry"}
                if updated:
                    print(key)
                    for attr in updated:
                        print("\t", attr, orig_dict[attr], parsed_data[attr])
        finally:
            # Ok to teardown region in the proxy now
            region.mark_dead()


addons = [ObjectManagementValidator()]




"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import unittest

from hippolyzer.lib.base.datatypes import *
from hippolyzer.lib.base.message.msgtypes import MsgType
from hippolyzer.lib.base.settings import Settings
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer
from hippolyzer.lib.base.message.data_packer import TemplateDataPacker


class TestSerializer(unittest.TestCase):

    def tearDown(self):
        pass

    def setUp(self):
        self.settings = Settings()
        self.settings.ENABLE_DEFERRED_PACKET_PARSING = False
        self.deserializer = UDPMessageDeserializer(settings=self.settings)
        self.serializer = UDPMessageSerializer()

    def test_serialize(self):
        body = b'\xff\xff\xff\xfb' + b'\x03' + \
               b'\x01\x00\x00\x00' + b'\x02\x00\x00\x00' + b'\x03\x00\x00\x00'
        message = b'\x00' + b'\x00\x00\x00\x01' + b'\x00' + body
        packet = self.deserializer.deserialize(message)
        packed_data = self.serializer.serialize(packet)
        assert packed_data == message, "Incorrect serialization"

    def test_acks_roundtrip(self):
        body = b'\xff\xff\xff\xfb' + b'\x03' + \
               b'\x01\x00\x00\x00' + b'\x02\x00\x00\x00' + b'\x03\x00\x00\x00'
        # Two acks
        trailer = b"\x00\x00\x00\x02\x00\x00\x00\x01\x02"
        message = b'\x10' + b'\x00\x00\x00\x01' + b'\x00' + body + trailer
        packet = self.deserializer.deserialize(message)
        self.assertEqual(packet.acks, (1, 2))
        self.assertEqual(self.serializer.serialize(packet), message)

    def test_template_data_packers(self):
        cases = {
            MsgType.MVT_FIXED: b"foobar",
            MsgType.MVT_VARIABLE: b"foobar\x00",
            MsgType.MVT_S8: -2,
            MsgType.MVT_U8: 2,
            MsgType.MVT_BOOL: 1,
            MsgType.MVT_LLUUID: UUID.random(),
            MsgType.MVT_IP_ADDR: "127.0.0.1",
            MsgType.MVT_IP_PORT: 500,
            MsgType.MVT_U16: 500,
            MsgType.MVT_U32: 500,
            MsgType.MVT_U64: 500,
            MsgType.MVT_S16: -500,
            MsgType.MVT_S32: -500,
            MsgType.MVT_S64: -500,
            MsgType.MVT_F32: 1.0,
            MsgType.MVT_F64: 1.0,
            MsgType.MVT_LLVector3: Vector3(),
            MsgType.MVT_LLVector3d: (0.0, 1.0, 0.0),
            MsgType.MVT_LLVector4: (0.0, 1.0, 0.0, 1.0),
            MsgType.MVT_LLQuaternion: Quaternion()
        }
        for case_type, case in cases.items():
            packed = TemplateDataPacker.pack(case, case_type)
            unpacked = TemplateDataPacker.unpack(packed, case_type)
            assert case == unpacked, (case_type, case, unpacked)

        # String case is special, it gets a null terminator.
        packed = TemplateDataPacker.pack("foobar", MsgType.MVT_VARIABLE)
        unpacked = TemplateDataPacker.unpack(packed, MsgType.MVT_VARIABLE)
        self.assertEqual(unpacked, b"foobar\x00")

    def test_simple_zero_coding(self):
        val = b"\x01\x00\x01"
        compressed = self.serializer.zero_code_compress(val)
        self.assertEqual(b"\x01\x00\x01\x01", compressed)

    def test_large_zero_coding(self):
        val = b"\x01" + (b"\x00" * 255) + b"\x00\x00\x01"
        compressed = self.serializer.zero_code_compress(val)
        # Make sure we didn't use the wrap case
        self.assertEqual(b"\x01\x00\xFF\x00\x02\x01", compressed)

    def test_twice_large_zero_coding(self):
        val = b"\x01" + (b"\x00" * 255 * 2) + b"\x00\x00\x01"
        compressed = self.serializer.zero_code_compress(val)
        self.assertEqual(b"\x01\x00\xFF\x00\xFF\x00\x02\x01", compressed)



"""
Demonstrates how addon state can be tied to sessions and survive
across reloads using the GlobalProperty and SessionProperty ClassVars
"""

from hippolyzer.lib.proxy.commands import handle_command, Parameter
from hippolyzer.lib.proxy.addon_utils import BaseAddon, SessionProperty, GlobalProperty, send_chat
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session


class AddonStateHelloWorldAddon(BaseAddon):
    # How to say hello, value shared across sessions and will be the same
    # regardless of which session is active when accessed.
    # "hello_greeting" is added to session_manager.addon_ctx's dict and will survive reloads
    # be mindful of conflicting with other addons' variables, other addons' variables
    # can be accessed manually through the session_manager.addon_ctx dict since
    # there isn't currently any namespacing.
    hello_greeting: str = GlobalProperty(default="Hello")
    # Who to say hello to, value specific to each session
    # Value will be different depending on which session is having its event
    # handled when the property is accessed.
    # "hello_person" is added to session.addon_ctx's dict and will survive reloads
    hello_person: str = SessionProperty(default="World")

    def __init__(self):
        # Tied to the addon instance.
        # Shared across sessions and will die if the addon is reloaded
        self.hello_punctuation = "!"

    @handle_command(
        # Use the longer-form `Parameter()` for declaring this because
        # this field should be greedy and take the rest of the message (no separator.)
        greeting=Parameter(str, sep=None),
    )
    async def set_hello_greeting(self, _session: Session, _region: ProxiedRegion, greeting: str):
        """Set the person to say hello to"""
        self.hello_greeting = greeting

    @handle_command(person=Parameter(str, sep=None))
    async def set_hello_person(self, _session: Session, _region: ProxiedRegion, person: str):
        """Set the person to say hello to"""
        self.hello_person = person

    @handle_command(
        # Punctuation should have no whitespace, so using a simple parameter is OK.
        punctuation=str,
    )
    async def set_hello_punctuation(self, _session: Session, _region: ProxiedRegion, punctuation: str):
        """Set the punctuation to use for saying hello"""
        self.hello_punctuation = punctuation

    @handle_command()
    async def say_hello(self, _session: Session, _region: ProxiedRegion):
        """Say hello using the configured hello variables"""
        # These aren't instance properties, they can be accessed via the class as well.
        hello_person = AddonStateHelloWorldAddon.hello_person
        send_chat(f"{self.hello_greeting} {hello_person}{self.hello_punctuation}")


addons = [AddonStateHelloWorldAddon()]
